/*
 * kgraft_patch_bsc1039496
 *
 * Fix for CVE-2017-1000364, bsc#1039496
 *
 *  SLE12(-SP1) commit:
 *   d14041c8b67b14a658572849b0ec7a9cb9c0acd3
 *
 *  SLE11-SP4 commit:
 *   cf4d269b082a02f96d205db88c6622ceb5908815
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if defined(CONFIG_X86_64)

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/binfmts.h>
#include <linux/highmem.h>
#include <linux/slab.h>
#include <linux/ptrace.h>
#include <linux/security.h>
#include <linux/fs_struct.h>
#include <linux/file.h>
#include <linux/sched.h>
#include <linux/seq_file.h>
#include <linux/khugepaged.h>
#include <linux/rmap.h>
#include <linux/mmu_notifier.h>
#include <linux/swap.h>
#include <linux/swapops.h>
#include <linux/delayacct.h>
#include <linux/hugetlb.h>
#include <asm/uaccess.h>
#include <asm/compat.h>
#include <asm/pgalloc.h>
#include <trace/events/sched.h>

#include "local.h"

static void (*kgr_putname)(struct filename *name);
static int (*kgr_unshare_files)(struct files_struct **displaced);
static int (*kgr_init_new_context)(struct task_struct *tsk, struct mm_struct *mm);
static void (*kgr_proc_exec_connector)(struct task_struct *task);
static void (*kgr_acct_update_integrals)(struct task_struct *tsk);
static void (*kgr_put_files_struct)(struct files_struct *files);
static void (*kgr_reset_files_struct)(struct files_struct *files);
static void (*kgr_ptrace_notify)(int exit_code);
static int (*kgr_insert_vm_struct)(struct mm_struct *mm, struct vm_area_struct *vma);
static struct kmem_cache **kgr_vm_area_cachep;
static void (*kgr_sched_exec)(void);
static int (*kgr_prepare_bprm_creds)(struct linux_binprm *bprm);
static void (*kgr_free_bprm)(struct linux_binprm *bprm);
static struct mm_struct *(*kgr_mm_alloc)(void);
struct tracepoint *kgr__tracepoint_sched_process_exec;
static const char *(*kgr_arch_vma_name)(struct vm_area_struct *vma);
static struct page *(*kgr_follow_page_mask)(struct vm_area_struct *vma, unsigned long address, unsigned int flags, unsigned int *page_mask);
static struct mm_struct *kgr_init_mm;
static int (*kgr_in_gate_area)(struct mm_struct *mm, unsigned long addr);
static struct page *(*kgr_vm_normal_page)(struct vm_area_struct *vma, unsigned long addr, pte_t pte);
static unsigned long *kgr_zero_pfn;
static struct vm_area_struct *(*kgr_get_gate_vma)(struct mm_struct *mm);
static int (*kgr_do_numa_page)(struct mm_struct *mm, struct vm_area_struct *vma, unsigned long addr, pte_t pte, pte_t *ptep, pmd_t *pmd);
static int (*kgr_do_wp_page)(struct mm_struct *mm, struct vm_area_struct *vma, unsigned long address, pte_t *page_table, pmd_t *pmd, spinlock_t *ptl, pte_t orig_pte);
static int (*kgr__do_fault)(struct mm_struct *mm, struct vm_area_struct *vma, unsigned long address, pmd_t *pmd, pgoff_t pgoff, unsigned int flags, pte_t orig_pte);
static void (*kgr_print_bad_pte)(struct vm_area_struct *vma, unsigned long addr, pte_t pte, struct page *page);
static int (*kgr__pud_alloc)(struct mm_struct *mm, pgd_t *pgd, unsigned long address);
static int (*kgr__pte_alloc)(struct mm_struct *mm, struct vm_area_struct *vma, pmd_t *pmd, unsigned long address);
static atomic_long_t *kgr_nr_swap_pages;
static long *kgr_total_swap_pages;
static bool (*kgr_mem_cgroup_oom_synchronize)(bool handle);
static struct page *(*kgr_lookup_swap_cache)(swp_entry_t entry);
static void (*kgr_page_add_new_anon_rmap)(struct page *page, struct vm_area_struct *vma, unsigned long address);
static int (*kgr_try_to_free_swap)(struct page *page);
static void (*kgr_swap_free)(swp_entry_t entry);
static void (*kgr_migration_entry_wait)(struct mm_struct *mm, pmd_t *pmd, unsigned long address);
static struct page *(*kgr_swapin_readahead)(swp_entry_t entry, gfp_t gfp_mask, struct vm_area_struct *vma, unsigned long addr);
static int (*kgr__lock_page_or_retry)(struct page *page, struct mm_struct *mm, unsigned int flags);
static int (*kgr_ptep_set_access_flags)(struct vm_area_struct *vma, unsigned long address, pte_t *ptep, pte_t entry, int dirty);
static int (*kgr_anon_vma_prepare)(struct vm_area_struct *vma);
static int (*kgr__pmd_alloc)(struct mm_struct *mm, pud_t *pud, unsigned long address);
static void (*kgr_do_page_add_anon_rmap)(struct page *page, struct vm_area_struct *vma, unsigned long address, int exclusive);
static struct vm_area_struct *(*kgr_find_extend_vma)(struct mm_struct *mm, unsigned long addr);
static int (*kgr_reuse_swap_page)(struct page *page);
static long (*kgr_vma_compute_subtree_gap)(struct vm_area_struct *vma);
static void (*kgr_perf_event_mmap)(struct vm_area_struct *vma);
static void (*kgr_anon_vma_interval_tree_remove)(struct anon_vma_chain *node, struct rb_root *root);
static int (*kgr_may_expand_vm)(struct mm_struct *mm, unsigned long npages);
static int (*kgr_security_vm_enough_memory_mm)(struct mm_struct *mm, long pages);
static void (*kgr_vm_stat_account)(struct mm_struct *mm, unsigned long flags, struct file *file, long pages);
static void (*kgr_anon_vma_interval_tree_insert)(struct anon_vma_chain *node, struct rb_root *root);
static int (*kgr_security_mmap_addr)(unsigned long addr);
static unsigned long *kgr_default_dump_filter;
static struct kmem_cache **kgr_mm_cachep;
static unsigned long *kgr_mmap_min_addr;
static pgd_t *(*kgr_pgd_alloc)(struct mm_struct *mm);
static void (*kgr_mm_init_owner)(struct mm_struct *mm, struct task_struct *p);
static void (*kgr_lru_cache_add_active_or_unevictable)(struct page *page, struct vm_area_struct *vma);
static void (*kgr_mem_cgroup_commit_charge)(struct page *page, struct mem_cgroup *memcg, bool lrucare);
static int (*kgr_mem_cgroup_try_charge)(struct page *page, struct mm_struct *mm, gfp_t gfp_mask, struct mem_cgroup **memcgp);
static void (*kgr_mem_cgroup_cancel_charge)(struct page *page, struct mem_cgroup *memcg);

#if defined(SPLIT_RSS_COUNTING)
static void (*kgr_add_mm_counter_fast)(struct mm_struct *mm, int member, int val);
static void (*kgr_sync_mm_rss)(struct mm_struct *mm);
#endif

#if !defined(CONFIG_XEN)
static pgtable_t (*kgr_pte_alloc_one)(struct mm_struct *mm, unsigned long address);
static struct page *(*kgr_get_huge_zero_page)(void);
static void (*kgr_put_huge_zero_page)(void);
static pmd_t (*kgr_maybe_pmd_mkwrite)(pmd_t pmd, struct vm_area_struct *vma);
static int (*kgr__khugepaged_enter)(struct mm_struct *mm);
static void (*kgr_clear_huge_page)(struct page *page, unsigned long addr, unsigned int pages_per_huge_page);
static void (*kgr_pgtable_trans_huge_deposit)(struct mm_struct *mm, pmd_t *pmdp, pgtable_t pgtable);
static unsigned long *kgr_transparent_hugepage_flags;
static spinlock_t *kgr_khugepaged_mm_lock;
static struct khugepaged_scan *kgr_khugepaged_scan;
static bool (*kgr_hugepage_vma_check)(struct vm_area_struct *vma);
static pmd_t *(*kgr_mm_find_pmd)(struct mm_struct *mm, unsigned long address);
static int (*kgr_khugepaged_node_load)[MAX_NUMNODES];
static unsigned int *kgr_khugepaged_max_ptes_none;
static int (*kgr_isolate_lru_page)(struct page *page);
static void (*kgr_putback_lru_page)(struct page *page);
static unsigned int *kgr_khugepaged_pages_collapsed;
static unsigned int *kgr_khugepaged_full_scans;
static void (*kgr_collect_mm_slot)(struct mm_slot *mm_slot);
static void (*kgr_free_page_and_swap_cache)(struct page *page);
static void (*kgr_page_remove_rmap)(struct page *page);
static pmd_t (*kgr_pmdp_clear_flush)(struct vm_area_struct *vma, unsigned long address, pmd_t *pmdp);
static int (*kgr__mmu_notifier_test_young)(struct mm_struct *mm, unsigned long address);
static int *kgr_zone_reclaim_mode;
static long (*kgr_follow_hugetlb_page)(struct mm_struct *mm, struct vm_area_struct *vma, struct page **pages, struct vm_area_struct **vmas, unsigned long *position, unsigned long *nr_pages, long i, unsigned int flags);
static int (*kgr_hugetlb_fault)(struct mm_struct *mm, struct vm_area_struct *vma, unsigned long address, unsigned int flags);
static bool (*kgr_is_vma_temporary_stack)(struct vm_area_struct *vma);
static int (*kgr_do_huge_pmd_numa_page)(struct mm_struct *mm, struct vm_area_struct *vma, unsigned long addr, pmd_t pmd, pmd_t *pmdp);
static int (*kgr_khugepaged_enter_vma_merge)(struct vm_area_struct *vma);
static void (*kgr_task_numa_free)(struct task_struct *p);
static struct page *(*kgr_ksm_might_need_to_copy)(struct page *page, struct vm_area_struct *vma, unsigned long address);
static struct page *(*kgr_alloc_pages_vma)(gfp_t gfp, int order, struct vm_area_struct *vma, unsigned long addr, int node, bool hugepage);
static int (*kgr_do_huge_pmd_wp_page)(struct mm_struct *mm, struct vm_area_struct *vma,	unsigned long address, pmd_t *pmd, pmd_t orig_pmd);
static void (*kgr_huge_pmd_set_accessed)(struct mm_struct *mm, struct vm_area_struct *vma, unsigned long address, pmd_t *pmd, pmd_t orig_pmd, int dirty);
static int *kgr_last_khugepaged_target_node;
static void (*kgr_pmd_clear_bad)(pmd_t *pmd);

#define kgr_alloc_hugepage_vma(gfp_mask, vma, addr, order)	\
	kgr_alloc_pages_vma(gfp_mask, order, vma, addr, numa_node_id(), true)
#define kgr_transparent_hugepage_enabled(__vma)				\
	((*kgr_transparent_hugepage_flags &					\
	  (1<<TRANSPARENT_HUGEPAGE_FLAG) ||				\
	  (*kgr_transparent_hugepage_flags &					\
	   (1<<TRANSPARENT_HUGEPAGE_REQ_MADV_FLAG) &&			\
	   ((__vma)->vm_flags & VM_HUGEPAGE))) &&			\
	 !((__vma)->vm_flags & VM_NOHUGEPAGE) &&			\
	 !kgr_is_vma_temporary_stack(__vma))
#define kgr_transparent_hugepage_defrag(__vma)				\
	((*kgr_transparent_hugepage_flags &					\
	  (1<<TRANSPARENT_HUGEPAGE_DEFRAG_FLAG)) ||			\
	 (*kgr_transparent_hugepage_flags &					\
	  (1<<TRANSPARENT_HUGEPAGE_DEFRAG_REQ_MADV_FLAG) &&		\
	  (__vma)->vm_flags & VM_HUGEPAGE))
#define kgr_transparent_hugepage_use_zero_page()				\
	(*kgr_transparent_hugepage_flags &					\
	 (1<<TRANSPARENT_HUGEPAGE_USE_ZERO_PAGE_FLAG))
#define kgr_khugepaged_always()				\
	(*kgr_transparent_hugepage_flags &			\
	 (1<<TRANSPARENT_HUGEPAGE_FLAG))
#define kgr_khugepaged_req_madv()					\
	(*kgr_transparent_hugepage_flags &				\
	 (1<<TRANSPARENT_HUGEPAGE_REQ_MADV_FLAG))
#define kgr_khugepaged_defrag()					\
	(*kgr_transparent_hugepage_flags &				\
	 (1<<TRANSPARENT_HUGEPAGE_DEFRAG_KHUGEPAGED_FLAG))

#define kgr_set_pte_at(mm, addr, ptep, pte)	native_set_pte_at(mm, addr, ptep, pte)

#else /* !CONFIG_XEN */

#define kgr_follow_hugetlb_page(m,v,p,vs,a,b,i,w)	({ BUG(); 0; })
#define kgr_hugetlb_fault(mm, vma, addr, flags)	({ BUG(); 0; })
#define kgr_transparent_hugepage_enabled(__vma) 0
#define kgr_alloc_pages_vma(gfp_mask, order, vma, addr, node, false)\
	alloc_pages(gfp_mask, order)

static inline void kgr_xen_set_pte_at(struct mm_struct *mm, unsigned long addr,
				  pte_t *ptep , pte_t pte)
{
	if ((mm != current->mm && mm != kgr_init_mm) ||
	    HYPERVISOR_update_va_mapping(addr, pte, 0))
		xen_set_pte(ptep, pte);
}

#define kgr_set_pte_at(mm, addr, ptep, pte)	kgr_xen_set_pte_at(mm, addr, ptep, pte)

static inline int kgr_do_huge_pmd_numa_page(struct mm_struct *mm, struct vm_area_struct *vma,
					unsigned long addr, pmd_t pmd, pmd_t *pmdp)
{
	return 0;
}
static inline int kgr_khugepaged_enter_vma_merge(struct vm_area_struct *vma)
{
	return 0;
}
static inline void kgr_task_numa_free(struct task_struct *p)
{
}
static inline struct page *kgr_ksm_might_need_to_copy(struct page *page,
			struct vm_area_struct *vma, unsigned long address)
{
	return page;
}


int kgr_do_huge_pmd_anonymous_page(struct mm_struct *mm,
	struct vm_area_struct *vma, unsigned long address, pmd_t *pmd,
	unsigned int flags);
static int (*kgr_do_huge_pmd_wp_page)(struct mm_struct *mm, struct vm_area_struct *vma,	unsigned long address, pmd_t *pmd, pmd_t orig_pmd);
static void (*kgr_huge_pmd_set_accessed)(struct mm_struct *mm, struct vm_area_struct *vma, unsigned long address, pmd_t *pmd, pmd_t orig_pmd, int dirty);
static void (*kgr_pmd_clear_bad)(pmd_t *pmd);

#endif /* !CONFIG_XEN */

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
#if !defined(CONFIG_XEN)
	{ "pte_alloc_one", (void *)&kgr_pte_alloc_one },
	{ "get_huge_zero_page", (void *)&kgr_get_huge_zero_page },
	{ "put_huge_zero_page", (void *)&kgr_put_huge_zero_page },
	{ "maybe_pmd_mkwrite", (void *)&kgr_maybe_pmd_mkwrite },
	{ "__khugepaged_enter", (void *)&kgr__khugepaged_enter },
	{ "clear_huge_page", (void *)&kgr_clear_huge_page },
	{ "pgtable_trans_huge_deposit", (void *)&kgr_pgtable_trans_huge_deposit },
	{ "transparent_hugepage_flags", (void *)&kgr_transparent_hugepage_flags },
	{ "khugepaged_mm_lock", (void *)&kgr_khugepaged_mm_lock },
	{ "khugepaged_scan", (void *)&kgr_khugepaged_scan },
	{ "hugepage_vma_check", (void *)&kgr_hugepage_vma_check },
	{ "mm_find_pmd", (void *)&kgr_mm_find_pmd },
	{ "khugepaged_node_load", (void *)&kgr_khugepaged_node_load },
	{ "khugepaged_max_ptes_none", (void *)&kgr_khugepaged_max_ptes_none },
	{ "isolate_lru_page", (void *)&kgr_isolate_lru_page },
	{ "putback_lru_page", (void *)&kgr_putback_lru_page },
	{ "khugepaged_pages_collapsed", (void *)&kgr_khugepaged_pages_collapsed },
	{ "khugepaged_full_scans", (void *)&kgr_khugepaged_full_scans },
	{ "collect_mm_slot", (void *)&kgr_collect_mm_slot },
	{ "free_page_and_swap_cache", (void *)&kgr_free_page_and_swap_cache },
	{ "page_remove_rmap", (void *)&kgr_page_remove_rmap },
	{ "pmdp_clear_flush", (void *)&kgr_pmdp_clear_flush },
	{ "__mmu_notifier_test_young", (void *)&kgr__mmu_notifier_test_young },
	{ "zone_reclaim_mode", (void *)&kgr_zone_reclaim_mode },
	{ "follow_hugetlb_page", (void *)&kgr_follow_hugetlb_page },
	{ "hugetlb_fault", (void *)&kgr_hugetlb_fault },
	{ "huge_pmd_set_accessed", (void *)&kgr_huge_pmd_set_accessed },
	{ "is_vma_temporary_stack", (void *)&kgr_is_vma_temporary_stack },
	{ "do_huge_pmd_numa_page", (void *)&kgr_do_huge_pmd_numa_page },
	{ "do_huge_pmd_wp_page", (void *)&kgr_do_huge_pmd_wp_page },
	{ "khugepaged_enter_vma_merge", (void *)&kgr_khugepaged_enter_vma_merge },
	{ "task_numa_free", (void *)&kgr_task_numa_free },
	{ "ksm_might_need_to_copy", (void *)&kgr_ksm_might_need_to_copy },
	{ "alloc_pages_vma", (void *)&kgr_alloc_pages_vma },
	{ "last_khugepaged_target_node.31268", (void *)&kgr_last_khugepaged_target_node },
	{ "pmd_clear_bad", (void *)&kgr_pmd_clear_bad },
#endif /* !CONFIG_XEN */
#if defined(SPLIT_RSS_COUNTING)
	{ "add_mm_counter_fast", (void *)&kgr_add_mm_counter_fast },
	{ "sync_mm_rss", (void *)&kgr_sync_mm_rss },
#endif
	{ "putname", (void *)&kgr_putname },
	{ "unshare_files", (void *)&kgr_unshare_files },
	{ "init_new_context", (void *)&kgr_init_new_context },
	{ "proc_exec_connector", (void *)&kgr_proc_exec_connector },
	{ "acct_update_integrals", (void *)&kgr_acct_update_integrals },
	{ "put_files_struct", (void *)&kgr_put_files_struct },
	{ "reset_files_struct", (void *)&kgr_reset_files_struct },
	{ "ptrace_notify", (void *)&kgr_ptrace_notify },
	{ "insert_vm_struct", (void *)&kgr_insert_vm_struct },
	{ "vm_area_cachep", (void *)&kgr_vm_area_cachep },
	{ "sched_exec", (void *)&kgr_sched_exec },
	{ "prepare_bprm_creds", (void *)&kgr_prepare_bprm_creds },
	{ "free_bprm", (void *)&kgr_free_bprm },
	{ "mm_alloc", (void *)&kgr_mm_alloc },
	{ "__tracepoint_sched_process_exec", (void *)&kgr__tracepoint_sched_process_exec },
	{ "arch_vma_name", (void *)&kgr_arch_vma_name },
	{ "follow_page_mask", (void *)&kgr_follow_page_mask },
	{ "init_mm", (void *)&kgr_init_mm },
	{ "in_gate_area", (void *)&kgr_in_gate_area },
	{ "vm_normal_page", (void *)&kgr_vm_normal_page },
	{ "zero_pfn", (void *)&kgr_zero_pfn },
	{ "get_gate_vma", (void *)&kgr_get_gate_vma },
	{ "do_numa_page", (void *)&kgr_do_numa_page },
	{ "do_wp_page", (void *)&kgr_do_wp_page },
	{ "__do_fault", (void *)&kgr__do_fault },
	{ "print_bad_pte", (void *)&kgr_print_bad_pte },
	{ "__pud_alloc", (void *)&kgr__pud_alloc },
	{ "__pte_alloc", (void *)&kgr__pte_alloc },
	{ "nr_swap_pages", (void *)&kgr_nr_swap_pages },
	{ "total_swap_pages", (void *)&kgr_total_swap_pages },
	{ "mem_cgroup_oom_synchronize", (void *)&kgr_mem_cgroup_oom_synchronize },
	{ "lookup_swap_cache", (void *)&kgr_lookup_swap_cache },
	{ "page_add_new_anon_rmap", (void *)&kgr_page_add_new_anon_rmap },
	{ "try_to_free_swap", (void *)&kgr_try_to_free_swap },
	{ "swap_free", (void *)&kgr_swap_free },
	{ "migration_entry_wait", (void *)&kgr_migration_entry_wait },
	{ "swapin_readahead", (void *)&kgr_swapin_readahead },
	{ "__lock_page_or_retry", (void *)&kgr__lock_page_or_retry },
	{ "ptep_set_access_flags", (void *)&kgr_ptep_set_access_flags },
	{ "anon_vma_prepare", (void *)&kgr_anon_vma_prepare },
	{ "__pmd_alloc", (void *)&kgr__pmd_alloc },
	{ "do_page_add_anon_rmap", (void *)&kgr_do_page_add_anon_rmap },
	{ "find_extend_vma", (void *)&kgr_find_extend_vma },
	{ "reuse_swap_page", (void *)&kgr_reuse_swap_page },
	{ "vma_compute_subtree_gap", (void *)&kgr_vma_compute_subtree_gap },
	{ "perf_event_mmap", (void *)&kgr_perf_event_mmap },
	{ "anon_vma_interval_tree_remove", (void *)&kgr_anon_vma_interval_tree_remove },
	{ "may_expand_vm", (void *)&kgr_may_expand_vm },
	{ "security_vm_enough_memory_mm", (void *)&kgr_security_vm_enough_memory_mm },
	{ "vm_stat_account", (void *)&kgr_vm_stat_account },
	{ "anon_vma_interval_tree_insert", (void *)&kgr_anon_vma_interval_tree_insert },
	{ "security_mmap_addr", (void *)&kgr_security_mmap_addr },
	{ "default_dump_filter", (void *)&kgr_default_dump_filter },
	{ "mm_cachep", (void *)&kgr_mm_cachep },
	{ "mmap_min_addr", (void *)&kgr_mmap_min_addr },
	{ "pgd_alloc", (void *)&kgr_pgd_alloc },
	{ "mm_init_owner", (void *)&kgr_mm_init_owner },
	{ "lru_cache_add_active_or_unevictable", (void *)&kgr_lru_cache_add_active_or_unevictable },
	{ "mem_cgroup_commit_charge", (void *)&kgr_mem_cgroup_commit_charge },
	{ "mem_cgroup_try_charge", (void *)&kgr_mem_cgroup_try_charge },
	{ "mem_cgroup_cancel_charge", (void *)&kgr_mem_cgroup_cancel_charge },
};

/* inlined */
static bool kgr_valid_arg_len(struct linux_binprm *bprm, long len)
{
	return len <= MAX_ARG_STRLEN;
}

/* inlined */
static const char __user *kgr_get_user_arg_ptr(struct user_arg_ptr argv, int nr)
{
	const char __user *native;

#ifdef CONFIG_COMPAT
	if (unlikely(argv.is_compat)) {
		compat_uptr_t compat;

		if (get_user(compat, argv.ptr.compat + nr))
			return ERR_PTR(-EFAULT);

		return compat_ptr(compat);
	}
#endif

	if (get_user(native, argv.ptr.native + nr))
		return ERR_PTR(-EFAULT);

	return native;
}

/*
 * The #ifdefs in the replacement function below haven't been taken
 * into account as CONFIG_STACK_GROWSUP is not set for x86_64 kernels.
 */
#if IS_ENABLED(CONFIG_STACK_GROWSUP)
#error "kGraft patch supports only !CONFIG_STACK_GROWSUP"
#endif

/* isra optimized */
static void kgr_acct_arg_size(struct linux_binprm *bprm, unsigned long pages)
{
	struct mm_struct *mm = current->mm;
	long diff = (long)(pages - bprm->vma_pages);

	if (!mm || !diff)
		return;

	bprm->vma_pages = pages;
	add_mm_counter(mm, MM_ANONPAGES, diff);
}

/*
 * Fix CVE-2017-1000364
 *  new MMF_ flag for new processes
 */
#define MMF_GUARD_GAP	21

/*
 * Fix CVE-2017-1000364
 *  new macro instead of stack_guard_gap
 */
#define STACK_GUARD_GAP(mm_flags)	(test_bit(MMF_GUARD_GAP, &mm_flags) ? \
		256UL<<PAGE_SHIFT : PAGE_SIZE)

/* inlined, fixed function */
static struct page *kgr_get_arg_page(struct linux_binprm *bprm,
	unsigned long pos, int write)
{
	struct page *page;
	int ret;

#ifdef CONFIG_STACK_GROWSUP
	if (write) {
		/*
		 * Fix CVE-2017-1000364
		 *  -ret = kgr_expand_downwards(bprm->vma, pos);
		 *  +1 line
		 */
		ret = kgr_expand_downwards(bprm->vma, pos, 0);
		if (ret < 0)
			return NULL;
	}
#endif
	ret = get_user_pages(current, bprm->mm, pos,
			1, write, 1, &page, NULL);
	if (ret <= 0)
		return NULL;

	if (write) {
		unsigned long size = bprm->vma->vm_end - bprm->vma->vm_start;
		struct rlimit *rlim;

		/*
		 * Fix CVE-2017-1000364
		 *  +1 line
		 */
		size -= STACK_GUARD_GAP(bprm->mm->flags);
		kgr_acct_arg_size(bprm, size / PAGE_SIZE);

		/*
		 * We've historically supported up to 32 pages (ARG_MAX)
		 * of argument strings even with small stacks
		 */
		if (size <= ARG_MAX)
			return page;

		/*
		 * Limit to 1/4-th the stack size for the argv+env strings.
		 * This ensures that:
		 *  - the remaining binfmt code will not run out of stack space,
		 *  - the program will have a reasonable amount of stack left
		 *    to work from.
		 */
		rlim = current->signal->rlim;
		if (size > ACCESS_ONCE(rlim[RLIMIT_STACK].rlim_cur) / 4) {
			put_page(page);
			return NULL;
		}
	}

	return page;
}

/* inlined */
static void kgr_put_arg_page(struct page *page)
{
	put_page(page);
}

/* inlined */
static void kgr_flush_arg_page(struct linux_binprm *bprm, unsigned long pos,
	struct page *page)
{
	flush_cache_page(bprm->vma, pos, page_to_pfn(page));
}

/* isra optimized */
static int kgr_copy_strings(int argc, struct user_arg_ptr argv,
	struct linux_binprm *bprm)
{
	struct page *kmapped_page = NULL;
	char *kaddr = NULL;
	unsigned long kpos = 0;
	int ret;

	while (argc-- > 0) {
		const char __user *str;
		int len;
		unsigned long pos;

		ret = -EFAULT;
		str = kgr_get_user_arg_ptr(argv, argc);
		if (IS_ERR(str))
			goto out;

		len = strnlen_user(str, MAX_ARG_STRLEN);
		if (!len)
			goto out;

		ret = -E2BIG;
		if (!kgr_valid_arg_len(bprm, len))
			goto out;

		/* We're going to work our way backwords. */
		pos = bprm->p;
		str += len;
		bprm->p -= len;

		while (len > 0) {
			int offset, bytes_to_copy;

			if (fatal_signal_pending(current)) {
				ret = -ERESTARTNOHAND;
				goto out;
			}
			cond_resched();

			offset = pos % PAGE_SIZE;
			if (offset == 0)
				offset = PAGE_SIZE;

			bytes_to_copy = offset;
			if (bytes_to_copy > len)
				bytes_to_copy = len;

			offset -= bytes_to_copy;
			pos -= bytes_to_copy;
			str -= bytes_to_copy;
			len -= bytes_to_copy;

			if (!kmapped_page || kpos != (pos & PAGE_MASK)) {
				struct page *page;

				page = kgr_get_arg_page(bprm, pos, 1);
				if (!page) {
					ret = -E2BIG;
					goto out;
				}

				if (kmapped_page) {
					flush_kernel_dcache_page(kmapped_page);
					kunmap(kmapped_page);
					kgr_put_arg_page(kmapped_page);
				}
				kmapped_page = page;
				kaddr = kmap(kmapped_page);
				kpos = pos & PAGE_MASK;
				kgr_flush_arg_page(bprm, kpos, kmapped_page);
			}
			if (copy_from_user(kaddr+offset, str, bytes_to_copy)) {
				ret = -EFAULT;
				goto out;
			}
		}
	}
	ret = 0;
out:
	if (kmapped_page) {
		flush_kernel_dcache_page(kmapped_page);
		kunmap(kmapped_page);
		kgr_put_arg_page(kmapped_page);
	}
	return ret;
}

/* patched function */
int kgr_copy_strings_kernel(int argc, const char *const *__argv,
			struct linux_binprm *bprm)
{
	int r;
	mm_segment_t oldfs = get_fs();
	struct user_arg_ptr argv = {
		.ptr.native = (const char __user *const  __user *)__argv,
	};

	set_fs(KERNEL_DS);
	r = kgr_copy_strings(argc, argv, bprm);
	set_fs(oldfs);

	return r;
}

/* inlined */
static int kgr_check_unsafe_exec(struct linux_binprm *bprm)
{
	struct task_struct *p = current, *t;
	unsigned n_fs;
	int res = 0;

	if (p->ptrace) {
		if (p->ptrace & PT_PTRACE_CAP)
			bprm->unsafe |= LSM_UNSAFE_PTRACE_CAP;
		else
			bprm->unsafe |= LSM_UNSAFE_PTRACE;
	}

	/*
	 * This isn't strictly necessary, but it makes it harder for LSMs to
	 * mess up.
	 */
	if (current->no_new_privs)
		bprm->unsafe |= LSM_UNSAFE_NO_NEW_PRIVS;

	n_fs = 1;
	spin_lock(&p->fs->lock);
	rcu_read_lock();
	for (t = next_thread(p); t != p; t = next_thread(t)) {
		if (t->fs == p->fs)
			n_fs++;
	}
	rcu_read_unlock();

	if (p->fs->users > n_fs) {
		bprm->unsafe |= LSM_UNSAFE_SHARE;
	} else {
		res = -EAGAIN;
		if (!p->fs->in_exec) {
			p->fs->in_exec = 1;
			res = 1;
		}
	}
	spin_unlock(&p->fs->lock);

	return res;
}

/* inlined */
static int kgr__bprm_mm_init(struct linux_binprm *bprm)
{
	int err;
	struct vm_area_struct *vma = NULL;
	struct mm_struct *mm = bprm->mm;

	bprm->vma = vma = kmem_cache_zalloc(*kgr_vm_area_cachep, GFP_KERNEL);
	if (!vma)
		return -ENOMEM;

	down_write(&mm->mmap_sem);
	vma->vm_mm = mm;

	/*
	 * Place the stack at the largest stack address the architecture
	 * supports. Later, we'll move this to an appropriate place. We don't
	 * use STACK_TOP because that can depend on attributes which aren't
	 * configured yet.
	 */
	BUILD_BUG_ON(VM_STACK_FLAGS & VM_STACK_INCOMPLETE_SETUP);
	vma->vm_end = STACK_TOP_MAX;
	vma->vm_start = vma->vm_end - PAGE_SIZE;
	vma->vm_flags = VM_SOFTDIRTY | VM_STACK_FLAGS | VM_STACK_INCOMPLETE_SETUP;
	vma->vm_page_prot = vm_get_page_prot(vma->vm_flags);
	INIT_LIST_HEAD(&vma->anon_vma_chain);

	err = kgr_insert_vm_struct(mm, vma);
	if (err)
		goto err;

	mm->stack_vm = mm->total_vm = 1;
	up_write(&mm->mmap_sem);
	bprm->p = vma->vm_end - sizeof(void *);
	return 0;
err:
	up_write(&mm->mmap_sem);
	bprm->vma = NULL;
	kmem_cache_free(*kgr_vm_area_cachep, vma);
	return err;
}

/* inlined */
static int kgr_bprm_mm_init(struct linux_binprm *bprm)
{
	int err;
	struct mm_struct *mm = NULL;

	bprm->mm = mm = kgr_mm_alloc();
	err = -ENOMEM;
	if (!mm)
		goto err;

	err = kgr_init_new_context(current, mm);
	if (err)
		goto err;

	err = kgr__bprm_mm_init(bprm);
	if (err)
		goto err;

	return 0;

err:
	if (mm) {
		bprm->mm = NULL;
		mmdrop(mm);
	}

	return err;
}

/* isra and constprop optimized */
static int kgr_count(struct user_arg_ptr argv, int max)
{
	int i = 0;

	if (argv.ptr.native != NULL) {
		for (;;) {
			const char __user *p = kgr_get_user_arg_ptr(argv, i);

			if (!p)
				break;

			if (IS_ERR(p))
				return -EFAULT;

			if (i >= max)
				return -E2BIG;
			++i;

			if (fatal_signal_pending(current))
				return -ERESTARTNOHAND;
			cond_resched();
		}
	}
	return i;
}

static inline void kgr_ptrace_event(int event, unsigned long message)
{
	if (unlikely(ptrace_event_enabled(current, event))) {
		current->ptrace_message = message;
		kgr_ptrace_notify((event << 8) | SIGTRAP);
	} else if (event == PTRACE_EVENT_EXEC) {
		/* legacy EXEC report via SIGTRAP */
		if ((current->ptrace & (PT_PTRACED|PT_SEIZED)) == PT_PTRACED)
			send_sig(SIGTRAP, current, 0);
	}
}

/* inlined */
static int kgr_exec_binprm(struct linux_binprm *bprm)
{
	pid_t old_pid, old_vpid;
	int ret;

	/* Need to fetch pid before load_binary changes it */
	old_pid = current->pid;
	rcu_read_lock();
	old_vpid = task_pid_nr_ns(current, task_active_pid_ns(current->parent));
	rcu_read_unlock();

	ret = search_binary_handler(bprm);
	if (ret >= 0) {
		kgr_trace_sched_process_exec(current, old_pid, bprm);
		kgr_ptrace_event(PTRACE_EVENT_EXEC, old_vpid);
		current->did_exec = 1;
		kgr_proc_exec_connector(current);

		if (bprm->file) {
			allow_write_access(bprm->file);
			fput(bprm->file);
			bprm->file = NULL; /* to catch use-after-free */
		}
	}

	return ret;
}

/* isra optimized */
static int kgr_do_execve_common(const char *filename,
	struct user_arg_ptr argv, struct user_arg_ptr envp)
{
	struct linux_binprm *bprm;
	struct file *file;
	struct files_struct *displaced;
	bool clear_in_exec;
	int retval;

	/*
	 * We move the actual failure in case of RLIMIT_NPROC excess from
	 * set*uid() to execve() because too many poorly written programs
	 * don't check setuid() return code.  Here we additionally recheck
	 * whether NPROC limit is still exceeded.
	 */
	if ((current->flags & PF_NPROC_EXCEEDED) &&
	    atomic_read(&current_user()->processes) > rlimit(RLIMIT_NPROC)) {
		retval = -EAGAIN;
		goto out_ret;
	}

	/* We're below the limit (still or again), so we don't want to make
	 * further execve() calls fail. */
	current->flags &= ~PF_NPROC_EXCEEDED;

	retval = kgr_unshare_files(&displaced);
	if (retval)
		goto out_ret;

	retval = -ENOMEM;
	bprm = kzalloc(sizeof(*bprm), GFP_KERNEL);
	if (!bprm)
		goto out_files;

	retval = kgr_prepare_bprm_creds(bprm);
	if (retval)
		goto out_free;

	retval = kgr_check_unsafe_exec(bprm);
	if (retval < 0)
		goto out_free;
	clear_in_exec = retval;
	current->in_execve = 1;

	file = open_exec(filename);
	retval = PTR_ERR(file);
	if (IS_ERR(file))
		goto out_unmark;

	kgr_sched_exec();

	bprm->file = file;
	bprm->filename = filename;
	bprm->interp = filename;

	retval = kgr_bprm_mm_init(bprm);
	if (retval)
		goto out_file;

	bprm->argc = kgr_count(argv, MAX_ARG_STRINGS);
	if ((retval = bprm->argc) < 0)
		goto out;

	bprm->envc = kgr_count(envp, MAX_ARG_STRINGS);
	if ((retval = bprm->envc) < 0)
		goto out;

	retval = prepare_binprm(bprm);
	if (retval < 0)
		goto out;

	retval = kgr_copy_strings_kernel(1, &bprm->filename, bprm);
	if (retval < 0)
		goto out;

	bprm->exec = bprm->p;
	retval = kgr_copy_strings(bprm->envc, envp, bprm);
	if (retval < 0)
		goto out;

	retval = kgr_copy_strings(bprm->argc, argv, bprm);
	if (retval < 0)
		goto out;

	retval = kgr_exec_binprm(bprm);
	if (retval < 0)
		goto out;

	/* execve succeeded */
	current->fs->in_exec = 0;
	current->in_execve = 0;
	kgr_acct_update_integrals(current);
	kgr_task_numa_free(current);
	kgr_free_bprm(bprm);
	if (displaced)
		kgr_put_files_struct(displaced);
	return retval;

out:
	if (bprm->mm) {
		kgr_acct_arg_size(bprm, 0);
		mmput(bprm->mm);
	}

out_file:
	if (bprm->file) {
		allow_write_access(bprm->file);
		fput(bprm->file);
	}

out_unmark:
	if (clear_in_exec)
		current->fs->in_exec = 0;
	current->in_execve = 0;

out_free:
	kgr_free_bprm(bprm);

out_files:
	if (displaced)
		kgr_reset_files_struct(displaced);
out_ret:
	return retval;
}

/* patched function */
int kgr_do_execve(const char *filename,
	const char __user *const __user *__argv,
	const char __user *const __user *__envp)
{
	struct user_arg_ptr argv = { .ptr.native = __argv };
	struct user_arg_ptr envp = { .ptr.native = __envp };

	return kgr_do_execve_common(filename, argv, envp);
}

/* patched function */
asmlinkage long kgr_sys_execve(const char __user * filename,
	const char __user *const __user *argv,
	const char __user *const __user *envp)
{
	struct filename *path = getname(filename);
	int error = PTR_ERR(path);

	if (!IS_ERR(path)) {
		error = kgr_do_execve(path->name, argv, envp);
		kgr_putname(path);
	}
	return error;
}

/* inlined */
static int kgr_compat_do_execve(const char *filename,
	const compat_uptr_t __user *__argv,
	const compat_uptr_t __user *__envp)
{
	struct user_arg_ptr argv = {
		.is_compat = true,
		.ptr.compat = __argv,
	};
	struct user_arg_ptr envp = {
		.is_compat = true,
		.ptr.compat = __envp,
	};
	return kgr_do_execve_common(filename, argv, envp);
}

/* patched function */
asmlinkage long kgr_compat_sys_execve(const char __user * filename,
	const compat_uptr_t __user * argv,
	const compat_uptr_t __user * envp)
{
	struct filename *path = getname(filename);
	int error = PTR_ERR(path);

	if (!IS_ERR(path)) {
		error = kgr_compat_do_execve(path->name, argv, envp);
		kgr_putname(path);
	}
	return error;
}

/* patched function */
int kgr_remove_arg_zero(struct linux_binprm *bprm)
{
	int ret = 0;
	unsigned long offset;
	char *kaddr;
	struct page *page;

	if (!bprm->argc)
		return 0;

	do {
		offset = bprm->p & ~PAGE_MASK;
		page = kgr_get_arg_page(bprm, bprm->p, 0);
		if (!page) {
			ret = -EFAULT;
			goto out;
		}
		kaddr = kmap_atomic(page);

		for (; offset < PAGE_SIZE && kaddr[offset];
				offset++, bprm->p++)
			;

		kunmap_atomic(kaddr);
		kgr_put_arg_page(page);

		/*
		 * kgr note: free_arg_page() is an empty function for CONFIG_MMU
		 *
		 * if (offset == PAGE_SIZE)
		 *	free_arg_page(bprm, (bprm->p >> PAGE_SHIFT) - 1);
		 */
	} while (offset == PAGE_SIZE);

	bprm->p++;
	bprm->argc--;
	ret = 0;

out:
	return ret;
}

/* inlined */
static void kgr_pad_len_spaces(struct seq_file *m, int len)
{
	len = 25 + sizeof(void*) * 6 - len;
	if (len < 1)
		len = 1;
	seq_printf(m, "%*c", len, ' ');
}

/*
 * Fix CVE-2017-1000364
 *  new function
 */
static int kgr_stack_guard_area(struct vm_area_struct *vma, unsigned long address)
{
	struct vm_area_struct *prev;

	if (!(vma->vm_flags & VM_GROWSDOWN))
		return 0;

	/*
	 * strictly speaking there is a guard gap between disjoint stacks
	 * but the gap is not canonical (it might be smaller) and it is
	 * reasonably safe to assume that we can ignore that gap for stack
	 * POPULATE or /proc/<pid>[s]maps purposes
	 */
	prev = vma->vm_prev;
	if (prev && prev->vm_flags & VM_GROWSDOWN)
		return 0;

	return address - vma->vm_start < STACK_GUARD_GAP(vma->vm_mm->flags);
}

/* patched function */
void kgr_show_map_vma(struct seq_file *m, struct vm_area_struct *vma)
{
	struct mm_struct *mm = vma->vm_mm;
	struct file *file = vma->vm_file;
	vm_flags_t flags = vma->vm_flags;
	unsigned long ino = 0;
	unsigned long long pgoff = 0;
	unsigned long start, end;
	dev_t dev = 0;
	int len;

	if (file) {
		struct inode *inode = file_inode(vma->vm_file);
		dev = inode_get_dev(inode);
		ino = inode->i_ino;
		pgoff = ((loff_t)vma->vm_pgoff) << PAGE_SHIFT;
	}

	/* We don't show the stack guard page in /proc/maps */
	start = vma->vm_start;
	/*
	 * Fix CVE-2017-1000364
	 *  -if (stack_guard_page_start(vma, start))
	 *  -	start += PAGE_SIZE;
	 */
	end = vma->vm_end;
	/*
	 * Fix CVE-2017-1000364
	 *  -if (stack_guard_page_end(vma, end))
	 *  -	end -= PAGE_SIZE;
	 *  +7 lines
	 */
	if (vma->vm_flags & VM_GROWSDOWN) {
		if (kgr_stack_guard_area(vma, start))
			start += STACK_GUARD_GAP(vma->vm_mm->flags);
	} else if (vma->vm_flags & VM_GROWSUP) {
		if (kgr_stack_guard_area(vma, end))
			end -= STACK_GUARD_GAP(vma->vm_mm->flags);
	}

	seq_printf(m, "%08lx-%08lx %c%c%c%c %08llx %02x:%02x %lu %n",
			start,
			end,
			flags & VM_READ ? 'r' : '-',
			flags & VM_WRITE ? 'w' : '-',
			flags & VM_EXEC ? 'x' : '-',
			flags & VM_MAYSHARE ? 's' : 'p',
			pgoff,
			MAJOR(dev), MINOR(dev), ino, &len);

	/*
	 * Print the dentry name for named mappings, and a
	 * special [heap] marker for the heap:
	 */
	if (file) {
		kgr_pad_len_spaces(m, len);
		seq_path(m, &file->f_path, "\n");
	} else {
		const char *name = kgr_arch_vma_name(vma);
		if (!name) {
			if (mm) {
				if (vma->vm_start <= mm->brk &&
						vma->vm_end >= mm->start_brk) {
					name = "[heap]";
				} else if (vma->vm_start <= mm->start_stack &&
					   vma->vm_end >= mm->start_stack) {
					name = "[stack]";
				}
			} else {
				name = "[vdso]";
			}
		}
		if (name) {
			kgr_pad_len_spaces(m, len);
			seq_puts(m, name);
		}
	}
	seq_putc(m, '\n');
}

#if !defined(CONFIG_XEN)
/* isra and part optimized */
static bool kgr_set_huge_zero_page(pgtable_t pgtable, struct mm_struct *mm,
	struct vm_area_struct *vma, unsigned long haddr, pmd_t *pmd,
	struct page *zero_page)
{
	pmd_t entry;
	if (!pmd_none(*pmd))
		return false;
	entry = mk_pmd(zero_page, vma->vm_page_prot);
	entry = pmd_wrprotect(entry);
	entry = pmd_mkhuge(entry);
	kgr_pgtable_trans_huge_deposit(mm, pmd, pgtable);
	set_pmd_at(mm, haddr, pmd, entry);
	atomic_long_inc(&mm->nr_ptes);
	return true;
}

/* inlined */
static inline gfp_t kgr_alloc_hugepage_gfpmask(int defrag, gfp_t extra_gfp)
{
	return (GFP_TRANSHUGE & ~(defrag ? 0 : __GFP_WAIT)) | extra_gfp;
}

/* inlined */
static inline pmd_t kgr_mk_huge_pmd(struct page *page, pgprot_t prot)
{
	pmd_t entry;
	entry = mk_pmd(page, prot);
	entry = pmd_mkhuge(entry);
	return entry;
}

/* inlined */
static int kgr__do_huge_pmd_anonymous_page(struct mm_struct *mm,
	struct vm_area_struct *vma, unsigned long haddr, pmd_t *pmd,
	struct page *page)
{
	struct mem_cgroup *memcg;
	pgtable_t pgtable;
	spinlock_t *ptl;

	VM_BUG_ON(!PageCompound(page));

	if (kgr_mem_cgroup_try_charge(page, mm, GFP_TRANSHUGE, &memcg))
		return VM_FAULT_OOM;

	pgtable = kgr_pte_alloc_one(mm, haddr);
	if (unlikely(!pgtable)) {
		kgr_mem_cgroup_cancel_charge(page, memcg);
		return VM_FAULT_OOM;
	}

	kgr_clear_huge_page(page, haddr, HPAGE_PMD_NR);
	/*
	 * The memory barrier inside __SetPageUptodate makes sure that
	 * clear_huge_page writes become visible before the set_pmd_at()
	 * write.
	 */
	__SetPageUptodate(page);

	ptl = pmd_lock(mm, pmd);
	if (unlikely(!pmd_none(*pmd))) {
		spin_unlock(ptl);
		kgr_mem_cgroup_cancel_charge(page, memcg);
		put_page(page);
		pte_free(mm, pgtable);
	} else {
		pmd_t entry;
		entry = kgr_mk_huge_pmd(page, vma->vm_page_prot);
		entry = kgr_maybe_pmd_mkwrite(pmd_mkdirty(entry), vma);
		kgr_page_add_new_anon_rmap(page, vma, haddr);
		kgr_mem_cgroup_commit_charge(page, memcg, false);
		kgr_lru_cache_add_active_or_unevictable(page, vma);
		kgr_pgtable_trans_huge_deposit(mm, pmd, pgtable);
		set_pmd_at(mm, haddr, pmd, entry);
		add_mm_counter(mm, MM_ANONPAGES, HPAGE_PMD_NR);
		atomic_long_inc(&mm->nr_ptes);
		spin_unlock(ptl);
	}

	return 0;
}

static inline int kgr_khugepaged_enter(struct vm_area_struct *vma)
{
	if (!test_bit(MMF_VM_HUGEPAGE, &vma->vm_mm->flags))
		if ((kgr_khugepaged_always() ||
		     (kgr_khugepaged_req_madv() &&
		      vma->vm_flags & VM_HUGEPAGE)) &&
		    !(vma->vm_flags & VM_NOHUGEPAGE))
			if (kgr__khugepaged_enter(vma->vm_mm))
				return -ENOMEM;
	return 0;
}

/* patched function */
int kgr_do_huge_pmd_anonymous_page(struct mm_struct *mm,
	struct vm_area_struct *vma, unsigned long address, pmd_t *pmd,
	unsigned int flags)
{
	gfp_t gfp;
	struct page *page;
	unsigned long haddr = address & HPAGE_PMD_MASK;

	if (haddr < vma->vm_start || haddr + HPAGE_PMD_SIZE > vma->vm_end)
		return VM_FAULT_FALLBACK;
	if (unlikely(kgr_anon_vma_prepare(vma)))
		return VM_FAULT_OOM;
	if (unlikely(kgr_khugepaged_enter(vma)))
		return VM_FAULT_OOM;
	/*
	 * Fix CVE-2017-1000364
	 *  +3 lines
	 */
	if (kgr_stack_guard_area(vma, haddr) ||
			kgr_stack_guard_area(vma, haddr + HPAGE_PMD_SIZE))
		return VM_FAULT_FALLBACK;
	if (!(flags & FAULT_FLAG_WRITE) &&
			kgr_transparent_hugepage_use_zero_page()) {
		spinlock_t *ptl;
		pgtable_t pgtable;
		struct page *zero_page;
		bool set;
		pgtable = kgr_pte_alloc_one(mm, haddr);
		if (unlikely(!pgtable))
			return VM_FAULT_OOM;
		zero_page = kgr_get_huge_zero_page();
		if (unlikely(!zero_page)) {
			pte_free(mm, pgtable);
			count_vm_event(THP_FAULT_FALLBACK);
			return VM_FAULT_FALLBACK;
		}
		ptl = pmd_lock(mm, pmd);
		set = kgr_set_huge_zero_page(pgtable, mm, vma, haddr, pmd,
				zero_page);
		spin_unlock(ptl);
		if (!set) {
			pte_free(mm, pgtable);
			kgr_put_huge_zero_page();
		}
		return 0;
	}
	gfp = kgr_alloc_hugepage_gfpmask(kgr_transparent_hugepage_defrag(vma), 0);
	page = kgr_alloc_hugepage_vma(gfp, vma, haddr, HPAGE_PMD_ORDER);
	if (unlikely(!page)) {
		count_vm_event(THP_FAULT_FALLBACK);
		return VM_FAULT_FALLBACK;
	}
	if (unlikely(kgr__do_huge_pmd_anonymous_page(mm, vma, haddr, pmd, page))) {
		put_page(page);
		count_vm_event(THP_FAULT_FALLBACK);
		return VM_FAULT_FALLBACK;
	}

	count_vm_event(THP_FAULT_ALLOC);
	return 0;
}

/* inlined */
static inline int kgr_khugepaged_test_exit(struct mm_struct *mm)
{
	return atomic_read(&mm->mm_users) == 0;
}

/* inlined */
static bool kgr_khugepaged_scan_abort(int nid)
{
	int i;

	/*
	 * If zone_reclaim_mode is disabled, then no extra effort is made to
	 * allocate memory locally.
	 */
	if (!*kgr_zone_reclaim_mode)
		return false;

	/* If there is a count for this node already, it must be acceptable */
	if ((*kgr_khugepaged_node_load)[nid])
		return false;

	for (i = 0; i < MAX_NUMNODES; i++) {
		if (!(*kgr_khugepaged_node_load)[i])
			continue;
		if (node_distance(nid, i) > RECLAIM_DISTANCE)
			return true;
	}
	return false;
}

/* inlined */
static int kgr_khugepaged_find_target_node(void)
{
	/*
	 * Fix CVE-2017-1000364
	 *  -static int last_khugepaged_target_node = NUMA_NO_NODE;
	 */
	int nid, target_node = 0, max_value = 0;

	/* find first node with max normal pages hit */
	for (nid = 0; nid < MAX_NUMNODES; nid++)
		if ((*kgr_khugepaged_node_load)[nid] > max_value) {
			max_value = (*kgr_khugepaged_node_load)[nid];
			target_node = nid;
		}

	/* do some balance if several nodes have the same hit record */
	if (target_node <= *kgr_last_khugepaged_target_node)
		for (nid = *kgr_last_khugepaged_target_node + 1; nid < MAX_NUMNODES;
				nid++)
			if (max_value == (*kgr_khugepaged_node_load)[nid]) {
				target_node = nid;
				break;
			}

	*kgr_last_khugepaged_target_node = target_node;
	return target_node;
}

/* inlined */
static struct page *kgr_khugepaged_alloc_page(struct page **hpage,
	struct mm_struct *mm, struct vm_area_struct *vma, unsigned long address,
	int node)
{
	gfp_t flags;

	VM_BUG_ON(*hpage);

	/* Only allocate from the target node */
	flags = kgr_alloc_hugepage_gfpmask(kgr_khugepaged_defrag(), __GFP_OTHER_NODE) |
	        __GFP_THISNODE;

	/*
	 * Before allocating the hugepage, release the mmap_sem read lock.
	 * The allocation can take potentially a long time if it involves
	 * sync compaction, and we do not need to hold the mmap_sem during
	 * that. We will recheck the vma after taking it again in write mode.
	 */
	up_read(&mm->mmap_sem);

	*hpage = alloc_pages_exact_node(node, flags, HPAGE_PMD_ORDER);
	if (unlikely(!*hpage)) {
		count_vm_event(THP_COLLAPSE_ALLOC_FAILED);
		*hpage = ERR_PTR(-ENOMEM);
		return NULL;
	}

	count_vm_event(THP_COLLAPSE_ALLOC);
	return *hpage;
}

/* inlined */
static void kgr_release_pte_page(struct page *page)
{
	/* 0 stands for page_is_file_cache(page) == false */
	dec_zone_page_state(page, NR_ISOLATED_ANON + 0);
	unlock_page(page);
	kgr_putback_lru_page(page);
}

/* inlined */
static void kgr_release_pte_pages(pte_t *pte, pte_t *_pte)
{
	while (--_pte >= pte) {
		pte_t pteval = *_pte;
		if (!pte_none(pteval))
			kgr_release_pte_page(pte_page(pteval));
	}
}

/* inlined */
static inline int kgr_mmu_notifier_test_young(struct mm_struct *mm,
					  unsigned long address)
{
	if (mm_has_notifiers(mm))
		return kgr__mmu_notifier_test_young(mm, address);
	return 0;
}

/* inlined */
static int kgr__collapse_huge_page_isolate(struct vm_area_struct *vma,
	unsigned long address, pte_t *pte)
{
	struct page *page;
	pte_t *_pte;
	int referenced = 0, none = 0;
	for (_pte = pte; _pte < pte+HPAGE_PMD_NR;
	     _pte++, address += PAGE_SIZE) {
		pte_t pteval = *_pte;
		if (pte_none(pteval)) {
			if (++none <= *kgr_khugepaged_max_ptes_none)
				continue;
			else
				goto out;
		}
		if (!pte_present(pteval) || !pte_write(pteval))
			goto out;
		page = kgr_vm_normal_page(vma, address, pteval);
		if (unlikely(!page))
			goto out;

		VM_BUG_ON(PageCompound(page));
		BUG_ON(!PageAnon(page));
		VM_BUG_ON(!PageSwapBacked(page));

		/* cannot use mapcount: can't collapse if there's a gup pin */
		if (page_count(page) != 1)
			goto out;
		/*
		 * We can do it before isolate_lru_page because the
		 * page can't be freed from under us. NOTE: PG_lock
		 * is needed to serialize against split_huge_page
		 * when invoked from the VM.
		 */
		if (!trylock_page(page))
			goto out;
		/*
		 * Isolate the page to avoid collapsing an hugepage
		 * currently in use by the VM.
		 */
		if (kgr_isolate_lru_page(page)) {
			unlock_page(page);
			goto out;
		}
		/* 0 stands for page_is_file_cache(page) == false */
		inc_zone_page_state(page, NR_ISOLATED_ANON + 0);
		VM_BUG_ON(!PageLocked(page));
		VM_BUG_ON(PageLRU(page));

		/* If there is no mapped pte young don't collapse the page */
		if (pte_young(pteval) || PageReferenced(page) ||
		    kgr_mmu_notifier_test_young(vma->vm_mm, address))
			referenced = 1;
	}
	if (likely(referenced))
		return 1;
out:
	kgr_release_pte_pages(pte, _pte);
	return 0;
}

/* inlined */
static void kgr__collapse_huge_page_copy(pte_t *pte, struct page *page,
	struct vm_area_struct *vma, unsigned long address, spinlock_t *ptl)
{
	pte_t *_pte;
	for (_pte = pte; _pte < pte+HPAGE_PMD_NR; _pte++) {
		pte_t pteval = *_pte;
		struct page *src_page;

		if (pte_none(pteval)) {
			clear_user_highpage(page, address);
			add_mm_counter(vma->vm_mm, MM_ANONPAGES, 1);
		} else {
			src_page = pte_page(pteval);
			copy_user_highpage(page, src_page, address, vma);
			VM_BUG_ON(page_mapcount(src_page) != 1);
			kgr_release_pte_page(src_page);
			/*
			 * ptl mostly unnecessary, but preempt has to
			 * be disabled to update the per-cpu stats
			 * inside page_remove_rmap().
			 */
			spin_lock(ptl);
			/*
			 * paravirt calls inside pte_clear here are
			 * superfluous.
			 */
			pte_clear(vma->vm_mm, address, _pte);
			kgr_page_remove_rmap(src_page);
			spin_unlock(ptl);
			kgr_free_page_and_swap_cache(src_page);
		}

		address += PAGE_SIZE;
		page++;
	}
}

/* inlined, fixed function */
static void kgr_collapse_huge_page(struct mm_struct *mm,
	unsigned long address, struct page **hpage, struct vm_area_struct *vma,
	int node)
{
	pmd_t *pmd, _pmd;
	pte_t *pte;
	pgtable_t pgtable;
	struct page *new_page;
	spinlock_t *pmd_ptl, *pte_ptl;
	int isolated;
	unsigned long hstart, hend;
	struct mem_cgroup *memcg;
	unsigned long mmun_start;	/* For mmu_notifiers */
	unsigned long mmun_end;		/* For mmu_notifiers */

	VM_BUG_ON(address & ~HPAGE_PMD_MASK);

	/* release the mmap_sem read lock. */
	new_page = kgr_khugepaged_alloc_page(hpage, mm, vma, address, node);
	if (!new_page)
		return;

	if (unlikely(kgr_mem_cgroup_try_charge(new_page, mm,
					   GFP_TRANSHUGE, &memcg)))
		return;

	/*
	 * Prevent all access to pagetables with the exception of
	 * gup_fast later hanlded by the ptep_clear_flush and the VM
	 * handled by the anon_vma lock + PG_lock.
	 */
	down_write(&mm->mmap_sem);
	if (unlikely(kgr_khugepaged_test_exit(mm)))
		goto out;

	vma = find_vma(mm, address);
	if (!vma)
		goto out;
	hstart = (vma->vm_start + ~HPAGE_PMD_MASK) & HPAGE_PMD_MASK;
	hend = vma->vm_end & HPAGE_PMD_MASK;
	if (address < hstart || address + HPAGE_PMD_SIZE > hend)
		goto out;
	if (!kgr_hugepage_vma_check(vma))
		goto out;
	/*
	 * Fix CVE-2017-1000364
	 *  +3 lines
	 */
	/* never try to collapse stack gap */
	if (kgr_stack_guard_area(vma, hstart) || kgr_stack_guard_area(vma, hend))
		goto out;
	pmd = kgr_mm_find_pmd(mm, address);
	if (!pmd)
		goto out;

	anon_vma_lock_write(vma->anon_vma);

	pte = pte_offset_map(pmd, address);
	pte_ptl = pte_lockptr(mm, pmd);

	mmun_start = address;
	mmun_end   = address + HPAGE_PMD_SIZE;
	mmu_notifier_invalidate_range_start(mm, mmun_start, mmun_end);
	pmd_ptl = pmd_lock(mm, pmd); /* probably unnecessary */
	/*
	 * After this gup_fast can't run anymore. This also removes
	 * any huge TLB entry from the CPU so we won't allow
	 * huge and small TLB entries for the same virtual address
	 * to avoid the risk of CPU bugs in that area.
	 */
	_pmd = kgr_pmdp_clear_flush(vma, address, pmd);
	spin_unlock(pmd_ptl);
	mmu_notifier_invalidate_range_end(mm, mmun_start, mmun_end);

	spin_lock(pte_ptl);
	isolated = kgr__collapse_huge_page_isolate(vma, address, pte);
	spin_unlock(pte_ptl);

	if (unlikely(!isolated)) {
		pte_unmap(pte);
		spin_lock(pmd_ptl);
		BUG_ON(!pmd_none(*pmd));
		/*
		 * We can only use set_pmd_at when establishing
		 * hugepmds and never for establishing regular pmds that
		 * points to regular pagetables. Use pmd_populate for that
		 */
		pmd_populate(mm, pmd, pmd_pgtable(_pmd));
		spin_unlock(pmd_ptl);
		anon_vma_unlock_write(vma->anon_vma);
		goto out;
	}

	/*
	 * All pages are isolated and locked so anon_vma rmap
	 * can't run anymore.
	 */
	anon_vma_unlock_write(vma->anon_vma);

	kgr__collapse_huge_page_copy(pte, new_page, vma, address, pte_ptl);
	pte_unmap(pte);
	__SetPageUptodate(new_page);
	pgtable = pmd_pgtable(_pmd);

	_pmd = kgr_mk_huge_pmd(new_page, vma->vm_page_prot);
	_pmd = kgr_maybe_pmd_mkwrite(pmd_mkdirty(_pmd), vma);

	/*
	 * spin_lock() below is not the equivalent of smp_wmb(), so
	 * this is needed to avoid the copy_huge_page writes to become
	 * visible after the set_pmd_at() write.
	 */
	smp_wmb();

	spin_lock(pmd_ptl);
	BUG_ON(!pmd_none(*pmd));
	kgr_page_add_new_anon_rmap(new_page, vma, address);
	kgr_mem_cgroup_commit_charge(new_page, memcg, false);
	kgr_lru_cache_add_active_or_unevictable(new_page, vma);
	kgr_pgtable_trans_huge_deposit(mm, pmd, pgtable);
	set_pmd_at(mm, address, pmd, _pmd);
	update_mmu_cache_pmd(vma, address, pmd);
	spin_unlock(pmd_ptl);

	*hpage = NULL;

	(*kgr_khugepaged_pages_collapsed)++;
out_up_write:
	up_write(&mm->mmap_sem);
	return;

out:
	kgr_mem_cgroup_cancel_charge(new_page, memcg);
	goto out_up_write;
}

/* inlined */
static int kgr_khugepaged_scan_pmd(struct mm_struct *mm,
	struct vm_area_struct *vma, unsigned long address,
	struct page **hpage)
{
	pmd_t *pmd;
	pte_t *pte, *_pte;
	int ret = 0, referenced = 0, none = 0;
	struct page *page;
	unsigned long _address;
	spinlock_t *ptl;
	int node = NUMA_NO_NODE;

	VM_BUG_ON(address & ~HPAGE_PMD_MASK);

	pmd = kgr_mm_find_pmd(mm, address);
	if (!pmd)
		goto out;

	memset(*kgr_khugepaged_node_load, 0, sizeof(*kgr_khugepaged_node_load));
	pte = pte_offset_map_lock(mm, pmd, address, &ptl);
	for (_address = address, _pte = pte; _pte < pte+HPAGE_PMD_NR;
	     _pte++, _address += PAGE_SIZE) {
		pte_t pteval = *_pte;
		if (pte_none(pteval)) {
			if (++none <= *kgr_khugepaged_max_ptes_none)
				continue;
			else
				goto out_unmap;
		}
		if (!pte_present(pteval) || !pte_write(pteval))
			goto out_unmap;
		page = kgr_vm_normal_page(vma, _address, pteval);
		if (unlikely(!page))
			goto out_unmap;
		/*
		 * Record which node the original page is from and save this
		 * information to khugepaged_node_load[].
		 * Khupaged will allocate hugepage from the node has the max
		 * hit record.
		 */
		node = page_to_nid(page);
		if (kgr_khugepaged_scan_abort(node))
			goto out_unmap;
		(*kgr_khugepaged_node_load)[node]++;
		VM_BUG_ON(PageCompound(page));
		if (!PageLRU(page) || PageLocked(page) || !PageAnon(page))
			goto out_unmap;
		/* cannot use mapcount: can't collapse if there's a gup pin */
		if (page_count(page) != 1)
			goto out_unmap;
		if (pte_young(pteval) || PageReferenced(page) ||
		    kgr_mmu_notifier_test_young(vma->vm_mm, address))
			referenced = 1;
	}
	if (referenced)
		ret = 1;
out_unmap:
	pte_unmap_unlock(pte, ptl);
	if (ret) {
		node = kgr_khugepaged_find_target_node();
		/* collapse_huge_page will return with the mmap_sem released */
		kgr_collapse_huge_page(mm, address, hpage, vma, node);
	}
out:
	return ret;
}

/* patched function */
unsigned int kgr_khugepaged_scan_mm_slot(unsigned int pages, struct page **hpage)
	__releases(kgr_khugepaged_mm_lock)
	__acquires(kgr_khugepaged_mm_lock)
{
	struct mm_slot *mm_slot;
	struct mm_struct *mm;
	struct vm_area_struct *vma;
	int progress = 0;

	VM_BUG_ON(!pages);
	VM_BUG_ON(NR_CPUS != 1 && !spin_is_locked(kgr_khugepaged_mm_lock));

	if (kgr_khugepaged_scan->mm_slot)
		mm_slot = kgr_khugepaged_scan->mm_slot;
	else {
		mm_slot = list_entry(kgr_khugepaged_scan->mm_head.next,
				     struct mm_slot, mm_node);
		kgr_khugepaged_scan->address = 0;
		kgr_khugepaged_scan->mm_slot = mm_slot;
	}
	spin_unlock(kgr_khugepaged_mm_lock);

	mm = mm_slot->mm;
	down_read(&mm->mmap_sem);
	if (unlikely(kgr_khugepaged_test_exit(mm)))
		vma = NULL;
	else
		vma = find_vma(mm, kgr_khugepaged_scan->address);

	progress++;
	for (; vma; vma = vma->vm_next) {
		unsigned long hstart, hend;

		cond_resched();
		if (unlikely(kgr_khugepaged_test_exit(mm))) {
			progress++;
			break;
		}
		if (!kgr_hugepage_vma_check(vma)) {
skip:
			progress++;
			continue;
		}
		hstart = (vma->vm_start + ~HPAGE_PMD_MASK) & HPAGE_PMD_MASK;
		hend = vma->vm_end & HPAGE_PMD_MASK;
		if (hstart >= hend)
			goto skip;
		if (kgr_khugepaged_scan->address > hend)
			goto skip;
		if (kgr_khugepaged_scan->address < hstart)
			kgr_khugepaged_scan->address = hstart;
		VM_BUG_ON(kgr_khugepaged_scan->address & ~HPAGE_PMD_MASK);

		while (kgr_khugepaged_scan->address < hend) {
			int ret;
			cond_resched();
			if (unlikely(kgr_khugepaged_test_exit(mm)))
				goto breakouterloop;

			VM_BUG_ON(kgr_khugepaged_scan->address < hstart ||
				  kgr_khugepaged_scan->address + HPAGE_PMD_SIZE >
				  hend);
			ret = kgr_khugepaged_scan_pmd(mm, vma,
						  kgr_khugepaged_scan->address,
						  hpage);
			/* move to next address */
			kgr_khugepaged_scan->address += HPAGE_PMD_SIZE;
			progress += HPAGE_PMD_NR;
			if (ret)
				/* we released mmap_sem so break loop */
				goto breakouterloop_mmap_sem;
			if (progress >= pages)
				goto breakouterloop;
		}
	}
breakouterloop:
	up_read(&mm->mmap_sem); /* exit_mmap will destroy ptes after this */
breakouterloop_mmap_sem:

	spin_lock(kgr_khugepaged_mm_lock);
	VM_BUG_ON(kgr_khugepaged_scan->mm_slot != mm_slot);
	/*
	 * Release the current mm_slot if this mm is about to die, or
	 * if we scanned all vmas of this mm.
	 */
	if (kgr_khugepaged_test_exit(mm) || !vma) {
		/*
		 * Make sure that if mm_users is reaching zero while
		 * khugepaged runs here, khugepaged_exit will find
		 * mm_slot not pointing to the exiting mm.
		 */
		if (mm_slot->mm_node.next != &kgr_khugepaged_scan->mm_head) {
			kgr_khugepaged_scan->mm_slot = list_entry(
				mm_slot->mm_node.next,
				struct mm_slot, mm_node);
			kgr_khugepaged_scan->address = 0;
		} else {
			kgr_khugepaged_scan->mm_slot = NULL;
			(*kgr_khugepaged_full_scans)++;
		}

		kgr_collect_mm_slot(mm_slot);
	}

	return progress;
}

#endif /* !CONFIG_XEN */

/* inlined */
static inline int kgr_is_zero_pfn(unsigned long pfn)
{
	return pfn == *kgr_zero_pfn;
}

static inline unsigned long kgr_my_zero_pfn(unsigned long addr)
{
	return *kgr_zero_pfn;
}

#if defined(SPLIT_RSS_COUNTING)
#define TASK_RSS_EVENTS_THRESH	(64)
static void kgr_check_sync_rss_stat(struct task_struct *task)
{
	if (unlikely(task != current))
		return;
	if (unlikely(task->rss_stat.events++ > TASK_RSS_EVENTS_THRESH))
		kgr_sync_mm_rss(task->mm);
}

#define kgr_inc_mm_counter_fast(mm, member) kgr_add_mm_counter_fast(mm, member, 1)
#define kgr_dec_mm_counter_fast(mm, member) kgr_add_mm_counter_fast(mm, member, -1)

#else /* SPLIT_RSS_COUNTING */
static void kgr_check_sync_rss_stat(struct task_struct *task)
{
}

#define kgr_inc_mm_counter_fast(mm, member) inc_mm_counter(mm, member)
#define kgr_dec_mm_counter_fast(mm, member) dec_mm_counter(mm, member)

#endif /* SPLIT_RSS_COUNTING */

/* inlined */
static int kgr_do_linear_fault(struct mm_struct *mm, struct vm_area_struct *vma,
	unsigned long address, pte_t *page_table, pmd_t *pmd,
	unsigned int flags, pte_t orig_pte)
{
	pgoff_t pgoff = (((address & PAGE_MASK)
			- vma->vm_start) >> PAGE_SHIFT) + vma->vm_pgoff;

	pte_unmap(page_table);
	/* The VMA was not fully populated on mmap() or missing VM_DONTEXPAND */
	if (!vma->vm_ops->fault)
		return VM_FAULT_SIGBUS;
	return kgr__do_fault(mm, vma, address, pmd, pgoff, flags, orig_pte);
}

/* inlined */
static inline struct page *
kgr_alloc_zeroed_user_highpage_movable(struct vm_area_struct *vma,
					unsigned long vaddr)
{
	return kgr__alloc_zeroed_user_highpage(__GFP_MOVABLE, vma, vaddr);
}

/* forward declaration */
int kgr_expand_stack(struct vm_area_struct *vma, unsigned long address);

/* inlined, fixed function */
static int kgr_do_anonymous_page(struct mm_struct *mm, struct vm_area_struct *vma,
	unsigned long address, pte_t *page_table, pmd_t *pmd, unsigned int flags)
{
	struct mem_cgroup *memcg;
	struct page *page;
	spinlock_t *ptl;
	pte_t entry;

	pte_unmap(page_table);

	/* File mapping without ->vm_ops ? */
	if (vma->vm_flags & VM_SHARED)
		return VM_FAULT_SIGBUS;

	/* Check if we need to add a guard page to the stack */
	/*
	 * Fix CVE-2017-1000364
	 *  -if (check_stack_guard_page(vma, address) < 0)
	 *  +2 lines
	 */
	if ((vma->vm_flags & (VM_GROWSDOWN|VM_GROWSUP)) &&
			kgr_expand_stack(vma, address) < 0)
		return VM_FAULT_SIGSEGV;

	/* Use the zero-page for reads */
	if (!(flags & FAULT_FLAG_WRITE)) {
		entry = pte_mkspecial(pfn_pte(kgr_my_zero_pfn(address),
						vma->vm_page_prot));
		page_table = pte_offset_map_lock(mm, pmd, address, &ptl);
		if (!pte_none(*page_table))
			goto unlock;
		goto setpte;
	}

	/* Allocate our own private page. */
	if (unlikely(kgr_anon_vma_prepare(vma)))
		goto oom;
	page = kgr_alloc_zeroed_user_highpage_movable(vma, address);
	if (!page)
		goto oom;

	if (kgr_mem_cgroup_try_charge(page, mm, GFP_KERNEL, &memcg))
		goto oom_free_page;

	/*
	 * The memory barrier inside __SetPageUptodate makes sure that
	 * preceeding stores to the page contents become visible before
	 * the set_pte_at() write.
	 */
	__SetPageUptodate(page);

	entry = mk_pte(page, vma->vm_page_prot);
	if (vma->vm_flags & VM_WRITE)
		entry = pte_mkwrite(pte_mkdirty(entry));

	page_table = pte_offset_map_lock(mm, pmd, address, &ptl);
	if (!pte_none(*page_table))
		goto release;

	kgr_inc_mm_counter_fast(mm, MM_ANONPAGES);
	kgr_page_add_new_anon_rmap(page, vma, address);
	kgr_mem_cgroup_commit_charge(page, memcg, false);
	kgr_lru_cache_add_active_or_unevictable(page, vma);
setpte:
	kgr_set_pte_at(mm, address, page_table, entry);

	/* No need to invalidate - it was non-present before */
	update_mmu_cache(vma, address, page_table);
unlock:
	pte_unmap_unlock(page_table, ptl);
	return 0;
release:
	kgr_mem_cgroup_cancel_charge(page, memcg);
	page_cache_release(page);
	goto unlock;
oom_free_page:
	page_cache_release(page);
oom:
	return VM_FAULT_OOM;
}

/* inlined */
static inline int kgr_pte_unmap_same(struct mm_struct *mm, pmd_t *pmd,
	pte_t *page_table, pte_t orig_pte)
{
	int same = 1;
#if defined(CONFIG_SMP) || defined(CONFIG_PREEMPT)
	if (sizeof(pte_t) > sizeof(unsigned long)) {
		spinlock_t *ptl = pte_lockptr(mm, pmd);
		spin_lock(ptl);
		same = pte_same(*page_table, orig_pte);
		spin_unlock(ptl);
	}
#endif
	pte_unmap(page_table);
	return same;
}

/* inlined */
static int kgr_do_nonlinear_fault(struct mm_struct *mm, struct vm_area_struct *vma,
	unsigned long address, pte_t *page_table, pmd_t *pmd, unsigned int flags,
	pte_t orig_pte)
{
	pgoff_t pgoff;

	flags |= FAULT_FLAG_NONLINEAR;

	if (!kgr_pte_unmap_same(mm, pmd, page_table, orig_pte))
		return 0;

	if (unlikely(!(vma->vm_flags & VM_NONLINEAR))) {
		/*
		 * Page table corrupted: show pte and kill process.
		 */
		kgr_print_bad_pte(vma, address, orig_pte, NULL);
		return VM_FAULT_SIGBUS;
	}

	pgoff = pte_to_pgoff(orig_pte);
	return kgr__do_fault(mm, vma, address, pmd, pgoff, flags, orig_pte);
}

/* inlined */
static inline bool kgr_vm_swap_full(void)
{
	return atomic_long_read(kgr_nr_swap_pages) * 2 < *kgr_total_swap_pages;
}

/* inlined */
static inline int kgr_lock_page_or_retry(struct page *page, struct mm_struct *mm,
				     unsigned int flags)
{
	might_sleep();
	return trylock_page(page) || kgr__lock_page_or_retry(page, mm, flags);
}

/* inlined */
static int kgr_do_swap_page(struct mm_struct *mm, struct vm_area_struct *vma,
	unsigned long address, pte_t *page_table, pmd_t *pmd,
	unsigned int flags, pte_t orig_pte)
{
	spinlock_t *ptl;
	struct page *page, *swapcache;
	struct mem_cgroup *memcg;
	swp_entry_t entry;
	pte_t pte;
	int locked;
	int exclusive = 0;
	int ret = 0;

	if (!kgr_pte_unmap_same(mm, pmd, page_table, orig_pte))
		goto out;

	entry = pte_to_swp_entry(orig_pte);
	if (unlikely(non_swap_entry(entry))) {
		if (is_migration_entry(entry)) {
			kgr_migration_entry_wait(mm, pmd, address);
		} else if (is_hwpoison_entry(entry)) {
			ret = VM_FAULT_HWPOISON;
		} else {
			kgr_print_bad_pte(vma, address, orig_pte, NULL);
			ret = VM_FAULT_SIGBUS;
		}
		goto out;
	}
	delayacct_set_flag(DELAYACCT_PF_SWAPIN);
	page = kgr_lookup_swap_cache(entry);
	if (!page) {
		page = kgr_swapin_readahead(entry,
					GFP_HIGHUSER_MOVABLE, vma, address);
		if (!page) {
			/*
			 * Back out if somebody else faulted in this pte
			 * while we released the pte lock.
			 */
			page_table = pte_offset_map_lock(mm, pmd, address, &ptl);
			if (likely(pte_same(*page_table, orig_pte)))
				ret = VM_FAULT_OOM;
			delayacct_clear_flag(DELAYACCT_PF_SWAPIN);
			goto unlock;
		}

		/* Had to read the page from swap area: Major fault */
		ret = VM_FAULT_MAJOR;
		count_vm_event(PGMAJFAULT);
		mem_cgroup_count_vm_event(mm, PGMAJFAULT);
	} else if (PageHWPoison(page)) {
		/*
		 * hwpoisoned dirty swapcache pages are kept for killing
		 * owner processes (which may be unknown at hwpoison time)
		 */
		ret = VM_FAULT_HWPOISON;
		delayacct_clear_flag(DELAYACCT_PF_SWAPIN);
		swapcache = page;
		goto out_release;
	}

	swapcache = page;
	locked = kgr_lock_page_or_retry(page, mm, flags);

	delayacct_clear_flag(DELAYACCT_PF_SWAPIN);
	if (!locked) {
		ret |= VM_FAULT_RETRY;
		goto out_release;
	}

	/*
	 * Make sure try_to_free_swap or reuse_swap_page or swapoff did not
	 * release the swapcache from under us.  The page pin, and pte_same
	 * test below, are not enough to exclude that.  Even if it is still
	 * swapcache, we need to check that the page's swap has not changed.
	 */
	if (unlikely(!PageSwapCache(page) || page_private(page) != entry.val))
		goto out_page;

	page = kgr_ksm_might_need_to_copy(page, vma, address);
	if (unlikely(!page)) {
		ret = VM_FAULT_OOM;
		page = swapcache;
		goto out_page;
	}

	if (kgr_mem_cgroup_try_charge(page, mm, GFP_KERNEL, &memcg)) {
		ret = VM_FAULT_OOM;
		goto out_page;
	}

	/*
	 * Back out if somebody else already faulted in this pte.
	 */
	page_table = pte_offset_map_lock(mm, pmd, address, &ptl);
	if (unlikely(!pte_same(*page_table, orig_pte)))
		goto out_nomap;

	if (unlikely(!PageUptodate(page))) {
		ret = VM_FAULT_SIGBUS;
		goto out_nomap;
	}

	/*
	 * The page isn't present yet, go ahead with the fault.
	 *
	 * Be careful about the sequence of operations here.
	 * To get its accounting right, reuse_swap_page() must be called
	 * while the page is counted on swap but not yet in mapcount i.e.
	 * before page_add_anon_rmap() and swap_free(); try_to_free_swap()
	 * must be called after the swap_free(), or it will never succeed.
	 */

	kgr_inc_mm_counter_fast(mm, MM_ANONPAGES);
	kgr_dec_mm_counter_fast(mm, MM_SWAPENTS);
	pte = mk_pte(page, vma->vm_page_prot);
	if ((flags & FAULT_FLAG_WRITE) && kgr_reuse_swap_page(page)) {
		pte = maybe_mkwrite(pte_mkdirty(pte), vma);
		flags &= ~FAULT_FLAG_WRITE;
		ret |= VM_FAULT_WRITE;
		exclusive = 1;
	}
	flush_icache_page(vma, page);
	if (pte_swp_soft_dirty(orig_pte))
		pte = pte_mksoft_dirty(pte);
	kgr_set_pte_at(mm, address, page_table, pte);
	if (page == swapcache) {
		kgr_do_page_add_anon_rmap(page, vma, address, exclusive);
		kgr_mem_cgroup_commit_charge(page, memcg, true);
	} else { /* ksm created a completely new copy */
		kgr_page_add_new_anon_rmap(page, vma, address);
		kgr_mem_cgroup_commit_charge(page, memcg, false);
		kgr_lru_cache_add_active_or_unevictable(page, vma);
	}

	kgr_swap_free(entry);
	if (kgr_vm_swap_full() || (vma->vm_flags & VM_LOCKED) || PageMlocked(page))
		kgr_try_to_free_swap(page);
	unlock_page(page);
	if (page != swapcache) {
		/*
		 * Hold the lock to avoid the swap entry to be reused
		 * until we take the PT lock for the pte_same() check
		 * (to avoid false positives from pte_same). For
		 * further safety release the lock after the swap_free
		 * so that the swap count won't change under a
		 * parallel locked swapcache.
		 */
		unlock_page(swapcache);
		page_cache_release(swapcache);
	}

	if (flags & FAULT_FLAG_WRITE) {
		ret |= kgr_do_wp_page(mm, vma, address, page_table, pmd, ptl, pte);
		if (ret & VM_FAULT_ERROR)
			ret &= VM_FAULT_ERROR;
		goto out;
	}

	/* No need to invalidate - it was non-present before */
	update_mmu_cache(vma, address, page_table);
unlock:
	pte_unmap_unlock(page_table, ptl);
out:
	return ret;
out_nomap:
	kgr_mem_cgroup_cancel_charge(page, memcg);
	pte_unmap_unlock(page_table, ptl);
out_page:
	unlock_page(page);
out_release:
	page_cache_release(page);
	if (page != swapcache) {
		unlock_page(swapcache);
		page_cache_release(swapcache);
	}
	return ret;
}

/* inlined */
static int kgr_handle_pte_fault(struct mm_struct *mm,
	struct vm_area_struct *vma, unsigned long address,
	pte_t *pte, pmd_t *pmd, unsigned int flags)
{
	pte_t entry;
	spinlock_t *ptl;

	entry = ACCESS_ONCE(*pte);
	if (!pte_present(entry)) {
		if (pte_none(entry)) {
			if (vma->vm_ops)
				return kgr_do_linear_fault(mm, vma, address,
						pte, pmd, flags, entry);
			return kgr_do_anonymous_page(mm, vma, address,
						 pte, pmd, flags);
		}
		if (pte_file(entry))
			return kgr_do_nonlinear_fault(mm, vma, address,
					pte, pmd, flags, entry);
		return kgr_do_swap_page(mm, vma, address,
					pte, pmd, flags, entry);
	}

	if (pte_protnone(entry))
		return kgr_do_numa_page(mm, vma, address, entry, pte, pmd);

	ptl = pte_lockptr(mm, pmd);
	spin_lock(ptl);
	if (unlikely(!pte_same(*pte, entry)))
		goto unlock;
	if (flags & FAULT_FLAG_WRITE) {
		if (!pte_write(entry))
			return kgr_do_wp_page(mm, vma, address,
					pte, pmd, ptl, entry);
		entry = pte_mkdirty(entry);
	}
	entry = pte_mkyoung(entry);
	if (kgr_ptep_set_access_flags(vma, address, pte, entry, flags & FAULT_FLAG_WRITE)) {
		update_mmu_cache(vma, address, pte);
	} else {
		/*
		 * This is needed only for protection faults but the arch code
		 * is not yet telling us if this is a protection fault or not.
		 * This still avoids useless tlb flushes for .text page faults
		 * with threads.
		 */
		if (flags & FAULT_FLAG_WRITE)
			flush_tlb_fix_spurious_fault(vma, address);
	}
unlock:
	pte_unmap_unlock(pte, ptl);
	return 0;
}

/*
 * pud_alloc() is defined in include/linux/mm.h
 */
#if defined(__ARCH_HAS_4LEVEL_HACK)
#error "kGraft patch supports only !__ARCH_HAS_4LEVEL_HACK"
#endif
/* inlined */
static inline pud_t *kgr_pud_alloc(struct mm_struct *mm, pgd_t *pgd, unsigned long address)
{
	return (unlikely(pgd_none(*pgd)) && kgr__pud_alloc(mm, pgd, address))?
		NULL: pud_offset(pgd, address);
}

static inline pmd_t *kgr_pmd_alloc(struct mm_struct *mm, pud_t *pud, unsigned long address)
{
	return (unlikely(pud_none(*pud)) && kgr__pmd_alloc(mm, pud, address))?
		NULL: pmd_offset(pud, address);
}

/* inlined */
static inline int kgr_pmd_none_or_trans_huge_or_clear_bad(pmd_t *pmd)
{
	pmd_t pmdval = pmd_read_atomic(pmd);
	/*
	 * The barrier will stabilize the pmdval in a register or on
	 * the stack so that it will stop changing under the code.
	 *
	 * When CONFIG_TRANSPARENT_HUGEPAGE=y on x86 32bit PAE,
	 * pmd_read_atomic is allowed to return a not atomic pmdval
	 * (for example pointing to an hugepage that has never been
	 * mapped in the pmd). The below checks will only care about
	 * the low part of the pmd with 32bit PAE x86 anyway, with the
	 * exception of pmd_none(). So the important thing is that if
	 * the low part of the pmd is found null, the high part will
	 * be also null or the pmd_none() check below would be
	 * confused.
	 */
#ifdef CONFIG_TRANSPARENT_HUGEPAGE
	barrier();
#endif
	if (pmd_none(pmdval) || pmd_trans_huge(pmdval))
		return 1;
	if (unlikely(pmd_bad(pmdval))) {
		kgr_pmd_clear_bad(pmd);
		return 1;
	}
	return 0;
}

/* inlined */
static inline int kgr_pmd_trans_unstable(pmd_t *pmd)
{
#ifdef CONFIG_TRANSPARENT_HUGEPAGE
	return kgr_pmd_none_or_trans_huge_or_clear_bad(pmd);
#else
	return 0;
#endif
}

/* inlined */
static int kgr__handle_mm_fault(struct mm_struct *mm,
	struct vm_area_struct *vma, unsigned long address, unsigned int flags)
{
	pgd_t *pgd;
	pud_t *pud;
	pmd_t *pmd;
	pte_t *pte;

	if (unlikely(is_vm_hugetlb_page(vma)))
		return kgr_hugetlb_fault(mm, vma, address, flags);

	pgd = pgd_offset(mm, address);
	pud = kgr_pud_alloc(mm, pgd, address);
	if (!pud)
		return VM_FAULT_OOM;
	pmd = kgr_pmd_alloc(mm, pud, address);
	if (!pmd)
		return VM_FAULT_OOM;
	if (pmd_none(*pmd) && kgr_transparent_hugepage_enabled(vma)) {
		int ret = VM_FAULT_FALLBACK;
		if (!vma->vm_ops)
			ret = kgr_do_huge_pmd_anonymous_page(mm, vma, address,
					pmd, flags);
		if (!(ret & VM_FAULT_FALLBACK))
			return ret;
	} else {
		pmd_t orig_pmd = *pmd;
		int ret;

		barrier();
		if (pmd_trans_huge(orig_pmd)) {
			unsigned int dirty = flags & FAULT_FLAG_WRITE;

			/*
			 * If the pmd is splitting, return and retry the
			 * the fault.  Alternative: wait until the split
			 * is done, and goto retry.
			 */
			if (pmd_trans_splitting(orig_pmd))
				return 0;

			if (pmd_protnone(orig_pmd))
				return kgr_do_huge_pmd_numa_page(mm, vma, address,
							     orig_pmd, pmd);

			if (dirty && !pmd_write(orig_pmd)) {
				ret = kgr_do_huge_pmd_wp_page(mm, vma, address, pmd,
							  orig_pmd);
				if (!(ret & VM_FAULT_FALLBACK))
					return ret;
			} else {
				kgr_huge_pmd_set_accessed(mm, vma, address, pmd,
						      orig_pmd, dirty);
				return 0;
			}
		}
	}

	/*
	 * Use __pte_alloc instead of pte_alloc_map, because we can't
	 * run pte_offset_map on the pmd, if an huge pmd could
	 * materialize from under us from a different thread.
	 */
	if (unlikely(pmd_none(*pmd)) &&
	    unlikely(kgr__pte_alloc(mm, vma, pmd, address)))
		return VM_FAULT_OOM;
	/*
	 * If a huge pmd materialized under us just retry later.  Use
	 * pmd_trans_unstable() instead of pmd_trans_huge() to ensure the pmd
	 * didn't become pmd_trans_huge under us and then back to pmd_none, as
	 * a result of MADV_DONTNEED running immediately after a huge pmd fault
	 * in a different thread of this mm, in turn leading to a misleading
	 * pmd_trans_huge() retval.  All we have to ensure is that it is a
	 * regular pmd that we can walk with pte_offset_map() and we can do that
	 * through an atomic read in C, which is what pmd_trans_unstable()
	 * provides.
	 */
	if (unlikely(kgr_pmd_trans_unstable(pmd)))
		return 0;
	/*
	 * A regular pmd is established and it can't morph into a huge pmd
	 * from under us anymore at this point because we hold the mmap_sem
	 * read mode and khugepaged takes it in write mode. So now it's
	 * safe to run pte_offset_map().
	 */
	pte = pte_offset_map(pmd, address);

	return kgr_handle_pte_fault(mm, vma, address, pte, pmd, flags);
}

/* patched function */
int kgr_handle_mm_fault(struct mm_struct *mm, struct vm_area_struct *vma,
	unsigned long address, unsigned int flags)
{
	int ret;

	__set_current_state(TASK_RUNNING);

	count_vm_event(PGFAULT);
	mem_cgroup_count_vm_event(mm, PGFAULT);

	/* do counter updates before entering really critical section. */
	kgr_check_sync_rss_stat(current);

	/*
	 * Enable the memcg OOM handling for faults triggered in user
	 * space.  Kernel faults are handled more gracefully.
	 */
	if (flags & FAULT_FLAG_USER)
		mem_cgroup_oom_enable();

	ret = kgr__handle_mm_fault(mm, vma, address, flags);

	if (flags & FAULT_FLAG_USER) {
		mem_cgroup_oom_disable();
                /*
                 * The task may have entered a memcg OOM situation but
                 * if the allocation error was handled gracefully (no
                 * VM_FAULT_OOM), there is no need to kill anything.
                 * Just clean up the OOM state peacefully.
                 */
                if (task_in_memcg_oom(current) && !(ret & VM_FAULT_OOM))
                        kgr_mem_cgroup_oom_synchronize(false);
	}

	return ret;
}

/* patched function */
long kgr__get_user_pages(struct task_struct *tsk, struct mm_struct *mm,
	unsigned long start, unsigned long nr_pages, unsigned int gup_flags,
	struct page **pages, struct vm_area_struct **vmas, int *nonblocking)
{
	long i;
	unsigned long vm_flags;
	unsigned int page_mask;

	if (!nr_pages)
		return 0;

	/*
	 * Require read or write permissions.
	 * If FOLL_FORCE is set, we only require the "MAY" flags.
	 */
	vm_flags  = (gup_flags & FOLL_WRITE) ?
			(VM_WRITE | VM_MAYWRITE) : (VM_READ | VM_MAYREAD);
	vm_flags &= (gup_flags & FOLL_FORCE) ?
			(VM_MAYREAD | VM_MAYWRITE) : (VM_READ | VM_WRITE);

	VM_BUG_ON(!!pages != !!(gup_flags & FOLL_GET));

	/*
	 * If FOLL_FORCE is set then do not force a full fault as the hinting
	 * fault information is unrelated to the reference behaviour of a task
	 * using the address space
	 */
	if (!(gup_flags & FOLL_FORCE))
		gup_flags |= FOLL_NUMA;

	i = 0;

	do {
		struct vm_area_struct *vma;

		vma = kgr_find_extend_vma(mm, start);
		if (!vma && kgr_in_gate_area(mm, start)) {
			unsigned long pg = start & PAGE_MASK;
			pgd_t *pgd;
			pud_t *pud;
			pmd_t *pmd;
			pte_t *pte;

			/* user gate pages are read-only */
			if (gup_flags & FOLL_WRITE)
				return i ? : -EFAULT;
			if (pg > TASK_SIZE)
				pgd = kgr_pgd_offset_k(pg);
			else
				pgd = pgd_offset_gate(mm, pg);
			BUG_ON(pgd_none(*pgd));
			pud = pud_offset(pgd, pg);
			BUG_ON(pud_none(*pud));
			pmd = pmd_offset(pud, pg);
			if (pmd_none(*pmd))
				return i ? : -EFAULT;
			VM_BUG_ON(pmd_trans_huge(*pmd));
			pte = pte_offset_map(pmd, pg);
			if (pte_none(*pte)) {
				pte_unmap(pte);
				return i ? : -EFAULT;
			}
			vma = kgr_get_gate_vma(mm);
			if (pages) {
				struct page *page;

				page = kgr_vm_normal_page(vma, start, *pte);
				if (!page) {
					if (!(gup_flags & FOLL_DUMP) &&
					     kgr_is_zero_pfn(pte_pfn(*pte)))
						page = pte_page(*pte);
					else {
						pte_unmap(pte);
						return i ? : -EFAULT;
					}
				}
				pages[i] = page;
				get_page(page);
			}
			pte_unmap(pte);
			page_mask = 0;
			goto next_page;
		}

#ifdef CONFIG_XEN
		if (vma && (vma->vm_flags & VM_FOREIGN)) {
			struct vm_foreign_map *foreign_map =
				vma->vm_private_data;
			struct page **map = foreign_map->map;
			int offset = (start - vma->vm_start) >> PAGE_SHIFT;
			if (map[offset] != NULL) {
			        if (pages) {
			                struct page *page = map[offset];

					pages[i] = page;
					get_page(page);
				}
				if (vmas)
					vmas[i] = vma;
				i++;
				start += PAGE_SIZE;
				nr_pages--;
				continue;
			}
		}
#endif
		if (!vma ||
		    (vma->vm_flags & (VM_IO | VM_PFNMAP)) ||
		    !(vm_flags & vma->vm_flags))
			return i ? : -EFAULT;

		if (is_vm_hugetlb_page(vma)) {
			i = kgr_follow_hugetlb_page(mm, vma, pages, vmas,
					&start, &nr_pages, i, gup_flags);
			continue;
		}

		do {
			struct page *page;
			unsigned int foll_flags = gup_flags;
			unsigned int page_increm;

			/*
			 * If we have a pending SIGKILL, don't keep faulting
			 * pages and potentially allocating memory.
			 */
			if (unlikely(fatal_signal_pending(current)))
				return i ? i : -ERESTARTSYS;

			cond_resched();
			while (!(page = kgr_follow_page_mask(vma, start,
						foll_flags, &page_mask))) {
				int ret;
				unsigned int fault_flags = 0;

				/* For mlock, just skip the stack guard page. */
				if (foll_flags & FOLL_MLOCK) {
					/*
					 * Fix CVE-2017-1000364
					 *  -if (stack_guard_page(vma, start))
					 *  +1 line
					 */
					if (kgr_stack_guard_area(vma, start))
						goto next_page;
				}
				if (foll_flags & FOLL_WRITE)
					fault_flags |= FAULT_FLAG_WRITE;
				if (nonblocking)
					fault_flags |= FAULT_FLAG_ALLOW_RETRY;
				if (foll_flags & FOLL_NOWAIT)
					fault_flags |= (FAULT_FLAG_ALLOW_RETRY | FAULT_FLAG_RETRY_NOWAIT);

				ret = kgr_handle_mm_fault(mm, vma, start,
							fault_flags);

				if (ret & VM_FAULT_ERROR) {
					if (ret & VM_FAULT_OOM)
						return i ? i : -ENOMEM;
					if (ret & (VM_FAULT_HWPOISON |
						   VM_FAULT_HWPOISON_LARGE)) {
						if (i)
							return i;
						else if (gup_flags & FOLL_HWPOISON)
							return -EHWPOISON;
						else
							return -EFAULT;
					}
					if (ret & (VM_FAULT_SIGBUS |
							VM_FAULT_SIGSEGV))
						return i ? i : -EFAULT;
					BUG();
				}

				if (tsk) {
					if (ret & VM_FAULT_MAJOR)
						tsk->maj_flt++;
					else
						tsk->min_flt++;
				}

				if (ret & VM_FAULT_RETRY) {
					if (nonblocking)
						*nonblocking = 0;
					return i;
				}

				/*
				 * The VM_FAULT_WRITE bit tells us that
				 * do_wp_page has broken COW when necessary,
				 * even if maybe_mkwrite decided not to set
				 * pte_write. We can thus safely do subsequent
				 * page lookups as if they were reads. But only
				 * do so when looping for pte_write is futile:
				 * in some cases userspace may also be wanting
				 * to write to the gotten user page, which a
				 * read fault here might prevent (a readonly
				 * page might get reCOWed by userspace write).
				 */
				if ((ret & VM_FAULT_WRITE) &&
				    !(vma->vm_flags & VM_WRITE))
					foll_flags |= FOLL_COW;

				cond_resched();
			}
			if (IS_ERR(page))
				return i ? i : PTR_ERR(page);
			if (pages) {
				pages[i] = page;

				flush_anon_page(vma, page, start);
				flush_dcache_page(page);
				page_mask = 0;
			}
next_page:
			if (vmas) {
				vmas[i] = vma;
				page_mask = 0;
			}
			page_increm = 1 + (~(start >> PAGE_SHIFT) & page_mask);
			if (page_increm > nr_pages)
				page_increm = nr_pages;
			i += page_increm;
			start += page_increm * PAGE_SIZE;
			nr_pages -= page_increm;
		} while (nr_pages && start < vma->vm_end);
	} while (nr_pages);
	return i;
}

/* inlined, fixed function */
/*
 * Fix CVE-2017-1000364
 *  prototype change
 */
static int kgr_acct_stack_growth(struct vm_area_struct *vma, unsigned long size,
	unsigned long grow, unsigned long gap)
{
	struct mm_struct *mm = vma->vm_mm;
	struct rlimit *rlim = current->signal->rlim;
	unsigned long new_start, actual_size;

	/* address space limit tests */
	if (!kgr_may_expand_vm(mm, grow))
		return -ENOMEM;

	/* Stack limit test */
	actual_size = size;
	if (size && (vma->vm_flags & (VM_GROWSUP | VM_GROWSDOWN)))
		/*
		 * Fix CVE-2017-1000364
		 *  -actual_size -= PAGE_SIZE;
		 *  +1 line
		 */
		actual_size -= gap;
	if (actual_size > ACCESS_ONCE(rlim[RLIMIT_STACK].rlim_cur))
		return -ENOMEM;

	/* mlock limit tests */
	if (vma->vm_flags & VM_LOCKED) {
		unsigned long locked;
		unsigned long limit;
		locked = mm->locked_vm + grow;
		limit = ACCESS_ONCE(rlim[RLIMIT_MEMLOCK].rlim_cur);
		limit >>= PAGE_SHIFT;
		if (locked > limit && !capable(CAP_IPC_LOCK))
			return -ENOMEM;
	}

	/* Check to ensure the stack will not grow into a hugetlb-only region */
	new_start = (vma->vm_flags & VM_GROWSUP) ? vma->vm_start :
			vma->vm_end - size;
	if (is_hugepage_only_range(vma->vm_mm, new_start, size))
		return -EFAULT;

	/*
	 * Overcommit..  This must be the final test, as it will
	 * update security statistics.
	 */
	if (kgr_security_vm_enough_memory_mm(mm, grow))
		return -ENOMEM;

	/* Ok, everything looks good - let it rip */
	if (vma->vm_flags & VM_LOCKED)
		mm->locked_vm += grow;
	kgr_vm_stat_account(mm, vma->vm_flags, vma->vm_file, grow);
	return 0;
}

/* inlined */
static inline void kgr_anon_vma_interval_tree_pre_update_vma(struct vm_area_struct *vma)
{
	struct anon_vma_chain *avc;

	list_for_each_entry(avc, &vma->anon_vma_chain, same_vma)
		kgr_anon_vma_interval_tree_remove(avc, &avc->anon_vma->rb_root);
}

/* inlined */
static inline void kgr_anon_vma_interval_tree_post_update_vma(struct vm_area_struct *vma)
{
	struct anon_vma_chain *avc;

	list_for_each_entry(avc, &vma->anon_vma_chain, same_vma)
		kgr_anon_vma_interval_tree_insert(avc, &avc->anon_vma->rb_root);
}

/* inlined */
static inline void kgr_vma_gap_callbacks_propagate(struct rb_node *rb, struct rb_node *stop)
{
	while (rb != stop) {
		struct vm_area_struct *node = rb_entry(rb, struct vm_area_struct, vm_rb);
		unsigned long augmented = kgr_vma_compute_subtree_gap(node);
		if (node->rb_subtree_gap == augmented)
			break;
		node->rb_subtree_gap = augmented;
		rb = rb_parent(&node->vm_rb);
	}
}

/* inlined */
static void kgr_vma_gap_update(struct vm_area_struct *vma)
{
	/*
	 * As it turns out, RB_DECLARE_CALLBACKS() already created a callback
	 * function that does exacltly what we want.
	 */
	kgr_vma_gap_callbacks_propagate(&vma->vm_rb, NULL);
}

/* patched function */
/*
 * Fix CVE-2017-1000364
 *  prototype change
 */
int kgr_expand_downwards(struct vm_area_struct *vma, unsigned long address,
	unsigned long gap)
{
	int error;

	/*
	 * We must make sure the anon_vma is allocated
	 * so that the anon_vma locking is not a noop.
	 */
	if (unlikely(kgr_anon_vma_prepare(vma)))
		return -ENOMEM;

	address &= PAGE_MASK;
	error = kgr_security_mmap_addr(address);
	if (error)
		return error;

	vma_lock_anon_vma(vma);

	/*
	 * vma->vm_start/vm_end cannot change under us because the caller
	 * is required to hold the mmap_sem in read mode.  We need the
	 * anon_vma lock to serialize against concurrent expand_stacks.
	 */

	/* Somebody else might have raced and expanded it already */
	if (address < vma->vm_start) {
		unsigned long size, grow;

		size = vma->vm_end - address;
		grow = (vma->vm_start - address) >> PAGE_SHIFT;

		error = -ENOMEM;
		if (grow <= vma->vm_pgoff) {
			/*
			 * Fix CVE-2017-1000364
			 *  -error = kgr_acct_stack_growth(vma, size, grow);
			 *  +1 line
			 */
			error = kgr_acct_stack_growth(vma, size, grow, gap);
			if (!error) {
				/*
				 * vma_gap_update() doesn't support concurrent
				 * updates, but we only hold a shared mmap_sem
				 * lock here, so we need to protect against
				 * concurrent vma expansions.
				 * vma_lock_anon_vma() doesn't help here, as
				 * we don't guarantee that all growable vmas
				 * in a mm share the same root anon vma.
				 * So, we reuse mm->page_table_lock to guard
				 * against concurrent vma expansions.
				 */
				spin_lock(&vma->vm_mm->page_table_lock);
				kgr_anon_vma_interval_tree_pre_update_vma(vma);
				vma->vm_start = address;
				vma->vm_pgoff -= grow;
				kgr_anon_vma_interval_tree_post_update_vma(vma);
				kgr_vma_gap_update(vma);
				spin_unlock(&vma->vm_mm->page_table_lock);

				kgr_perf_event_mmap(vma);
			}
		}
	}
	vma_unlock_anon_vma(vma);
	kgr_khugepaged_enter_vma_merge(vma);
	kgr_validate_mm(vma->vm_mm);
	return error;
}

/*
 * Fix CVE-2017-1000364
 *  +new function
 */
unsigned long kgr_expandable_stack_area(struct vm_area_struct *vma,
	unsigned long address, unsigned long *gap)
{
	struct vm_area_struct *prev = vma->vm_prev;
	unsigned long guard_gap = STACK_GUARD_GAP(vma->vm_mm->flags);
	unsigned long guard_addr;

	address &= PAGE_MASK;
	if (!prev)
		goto out;

	/*
	* Is there a mapping abutting this one below?
	*
	* That's only ok if it's the same stack mapping
	* that has gotten split or there is sufficient gap
	* between mappings
	*
	* Fix CVE-2017-1000364
	*  +comment, from a follow-up fix
	*
	* Please note that some application (e.g. Java) punches
	* MAP_FIXED inside the stack and then PROT_NONE it
	* to mimic a stack guard which will clash with our protection
	* so pretend tha PROT_NONE vmas are OK
	*
	*/
	/*
	 * Fix CVE-2017-1000364
	 *  follow-up fix
	 *  -if (prev->vm_flags & VM_GROWSDOWN) {
	 *  +1 line
	 */
	if ((prev->vm_flags & VM_GROWSDOWN) || !(prev->vm_flags & (VM_WRITE|VM_READ))) {
		guard_gap = min(guard_gap, address - prev->vm_end);
		goto out;
	}

	if (address - prev->vm_end < guard_gap)
		return -ENOMEM;

out:
	/* make sure we won't underflow */
	if (address < *kgr_mmap_min_addr)
		return -ENOMEM;
	if (address - *kgr_mmap_min_addr < guard_gap)
		guard_gap = address - *kgr_mmap_min_addr;

	guard_addr = address - guard_gap;
	*gap = guard_gap;

	return guard_addr;
}

/* patched function */
/*
 * Fix CVE-2017-1000364
 *  function rewritten
 */
int kgr_expand_stack(struct vm_area_struct *vma, unsigned long address)
{
	unsigned long gap;

	address = kgr_expandable_stack_area(vma, address, &gap);
	if (IS_ERR_VALUE(address))
		return -ENOMEM;
	return kgr_expand_downwards(vma, address, gap);
}

/* inlined */
static void kgr_mm_init_aio(struct mm_struct *mm)
{
#ifdef CONFIG_AIO
	spin_lock_init(&mm->ioctx_lock);
	mm->ioctx_table = NULL;
#endif
}

/* inlined */
static inline int kgr_mm_alloc_pgd(struct mm_struct *mm)
{
	mm->pgd = kgr_pgd_alloc(mm);
	if (unlikely(!mm->pgd))
		return -ENOMEM;
	return 0;
}

/* patched function */
struct mm_struct *kgr_mm_init(struct mm_struct *mm, struct task_struct *p)
{
	atomic_set(&mm->mm_users, 1);
	atomic_set(&mm->mm_count, 1);
	init_rwsem(&mm->mmap_sem);
	INIT_LIST_HEAD(&mm->mmlist);
	mm->core_state = NULL;
	atomic_long_set(&mm->nr_ptes, 0);
	memset(&mm->rss_stat, 0, sizeof(mm->rss_stat));
	spin_lock_init(&mm->page_table_lock);
	kgr_mm_init_aio(mm);
	kgr_mm_init_owner(mm, p);
	clear_tlb_flush_pending(mm);

	if (current->mm) {
		mm->flags = current->mm->flags & MMF_INIT_MASK;
		mm->def_flags = current->mm->def_flags & VM_INIT_DEF_MASK;
	} else {
		mm->flags = *kgr_default_dump_filter;
		mm->def_flags = 0;
	}

	/*
	 * Fix CVE-2017-1000364
	 *  set new MMF_ for new processes
	 */
	set_bit(MMF_GUARD_GAP, &mm->flags);

	if (likely(!kgr_mm_alloc_pgd(mm))) {
		mmu_notifier_mm_init(mm);
		return mm;
	}

	kgr_free_mm(mm);
	return NULL;
}

static int kgr_patch_bsc1039496_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

int kgr_patch_bsc1039496_init(void)
{
	return kgr_patch_bsc1039496_kallsyms();
}

#endif /* CONFIG_X86_64 */
